#  Copyright (c) 2020, Build-A-Cell. All rights reserved.
#  See LICENSE file in the project root directory for details.

from biocrnpyler import CombinatorialComplex, One_Step_Binding, Reaction, CombinatorialConformation
from biocrnpyler import Species, OrderedPolymerSpecies, Complex, PolymerConformation
import pytest

def test_CombinatorialComplex_init_and_properties():
    X, Y, Z = Species("X"), Species("Y"), Species("Z")
    
    #Single final_state case
    C1 = Complex([Y, Z, Z])
    CC1 = CombinatorialComplex(final_states = [C1])

    #tests getters and setters
    assert set(CC1.final_states) == set([C1])
    assert set(CC1.sub_species) == set([Y, Z])
    assert set(CC1.initial_states) == set([Y, Z])
    assert CC1.intermediate_states is None

    #multiple final_state case
    C2 = Complex([X, X, Z, Z])
    CC2 = CombinatorialComplex(final_states = [C1, C2])

    #tests getters and setters
    assert set(CC2.final_states) == set([C1, C2])
    assert set(CC2.sub_species) == set([X, Y, Z])
    assert set(CC2.initial_states) == set([X, Y, Z])
    assert CC2.intermediate_states is None

    #test with initial states
    CC3 = CombinatorialComplex(final_states = [C1, C2], initial_states = [Complex([Z, Z]), X])
    assert set(CC3.final_states) == set([C1, C2])
    assert set(CC3.sub_species) == set([X, Y, Z])
    assert set(CC3.initial_states) == set([Complex([Z, Z]), X])
    assert CC3.intermediate_states is None

    #Test with intermediate states
    CC4 = CombinatorialComplex(final_states = [C1, C2], intermediate_states = [Complex([Z, Z])])
    assert set(CC4.final_states) == set([C1, C2])
    assert set(CC4.sub_species) == set([X, Y, Z])
    assert set(CC4.initial_states) == set([X, Y, Z])
    assert set(CC4.intermediate_states) == set([Complex([Z, Z])])

    #Test with excluded states
    CC5 = CombinatorialComplex(final_states = [C2], excluded_states = [C1])

    #Test cases that should produce errors

    #final_states must be ComplexSpecies
    with pytest.raises(ValueError):
        CC = CombinatorialComplex(final_states = [Species("C1")])

    #initial_states must be Species inside of final states
    with pytest.raises(ValueError):
        CC = CombinatorialComplex(final_states = [C1], initial_states = [Species("S")])

    #initial states which are Complexes can only contain things in the final_states
    with pytest.raises(ValueError):
        CC = CombinatorialComplex(final_states = [C1], initial_states = [Complex([Species("S"), Species("S2")])])

    #intermiade_states must be Complexes
    with pytest.raises(ValueError):
        CC = CombinatorialComplex(final_states = [C1], intermediate_states = [Species("S")])

    #intermediate_states must contain only things in final states
    with pytest.raises(ValueError):
        CC = CombinatorialComplex(final_states = [C1], intermediate_states = [Complex([Species("S"), Species("S2")])])

def test_CombinatorialComplex_compute_species_to_add():
    X, Y, Z = Species("X"), Species("Y"), Species("Z")
    C = Complex([X, X, Y, Z])
    CC = CombinatorialComplex(final_states = [C])

    #A single X, Y, and Z should be added
    species_to_add = CC.compute_species_to_add(X, C)
    assert set(species_to_add) == set([X, Y, Z])
    assert species_to_add.count(X) == 1 and species_to_add.count(Y) == 1 and species_to_add.count(Z) == 1

    #Two X's, and Z should be added
    species_to_add = CC.compute_species_to_add(Y, C)
    assert set(species_to_add) == set([X, Z])
    assert species_to_add.count(X) == 2 and species_to_add.count(Z) == 1

    #Try with s0 as a ComplexSpecies
    species_to_add = CC.compute_species_to_add(Complex([X, X]), C)
    assert set(species_to_add) == set([Y, Z])
    assert species_to_add.count(Y) == 1 and species_to_add.count(Z) == 1

    #If sf is not a ComplexSpecies, there is an error
    with pytest.raises(ValueError):
        species_to_add = CC.compute_species_to_add(C, X)

    #In the following cases, sf cannot be created by adding species to s0, so None is returned

    #C contains more species than Complex([X, X])
    species_to_add = CC.compute_species_to_add(C, Complex([X, Y]))
    assert species_to_add is None
    #C contains more species than Complex([X, X])
    species_to_add = CC.compute_species_to_add(C, Complex([X, X]))
    assert species_to_add is None
    #S is not in C
    species_to_add = CC.compute_species_to_add(Species("S"), C)
    assert species_to_add is None
    species_to_add = CC.compute_species_to_add(Complex([Species("S"), X]), C)
    assert species_to_add is None

def test_CombinatorialComplex_get_combinations_between():
    X, Y, Z = Species("X"), Species("Y"), Species("Z")
    C = Complex([X, Y, Z])
    CC = CombinatorialComplex(final_states = [C])

    combos = CC.get_combinations_between(X, C)

    #Below conditions uniquely specify all the correct returns
    assert len(combos) == 4
    #Assert first steps are in combos
    assert (Y, X, Complex([X, Y])) in combos and (Z, X, Complex([X, Z])) in combos
    #Second steps are in combos
    assert (Z, Complex([X, Y]), Complex([X, Y, Z])) in combos and (Y, Complex([X, Z]), Complex([X, Y, Z])) in combos

    #test again from a complex
    combos = CC.get_combinations_between(Complex([X, Y]), C)
    assert len(combos) == 1
    assert (Z, Complex([X, Y]), Complex([X, Y, Z])) in combos


    #Test with duplicate species
    C2 = Complex([X, X, Z])
    CC2 = CombinatorialComplex(final_states = [C2])
    combos = CC2.get_combinations_between(X, C2)
    assert len(combos) == 4
    assert (X, X, Complex([X, X])) in combos and (Z, X, Complex([X, Z])) in combos
    assert (Z, Complex([X, X]), C2) in combos and (X, Complex([X, Z]), C2) in combos

    #test again from a Complex
    combos = CC2.get_combinations_between(Complex([X, X]), C2)
    assert len(combos) == 1
    assert (Z, Complex([X, X]), C2) in combos
    combos = CC2.get_combinations_between(Complex([X, Z]), C2)
    assert len(combos) == 1
    assert (X, Complex([X, Z]), C2) in combos

    #Test with excluded states that are ComplexSpecies
    C3a = Complex([X, Y])
    C3b = Complex([X, Z])
    CC3 = CombinatorialComplex(final_states = [C], excluded_states = [C3a, C3b])
    combos = CC3.get_combinations_between(X, C)
    combos_list = [c[0] for c in combos]+[c[1] for c in combos] + [c[2] for c in combos]
    assert C3a not in combos and C3b not in combos

    #Test with excluded states that are Species
    CC4 = CombinatorialComplex(final_states = [C], excluded_states = [Y, Z])
    combos = CC4.get_combinations_between(X, C)
    combos_list = [c[0] for c in combos]+[c[1] for c in combos] + [c[2] for c in combos]
    assert Y not in combos and Z not in combos 

def test_CombinatorialComplex_update_species():
    mech_b = One_Step_Binding()

    #Basic Case
    X, Y, Z = Species("X"), Species("Y"), Species("Z")
    C1 = Complex([X, Y, Z])
    CC1 = CombinatorialComplex(final_states = [C1], mechanisms = [mech_b])

    s1 = CC1.update_species()
    assert X in s1 and Y in s1 and Z in s1 and C1 in s1
    assert Complex([X, Y]) in s1 and Complex([X, Z]) in s1 and Complex([Y, Z]) in s1

    #Set initial states
    C2i1 = Complex([X, X, Y])
    C2i2 = Complex([Y, Z])
    C2 = Complex([X, X, Y, Z])
    CC2 = CombinatorialComplex(final_states = [C2], initial_states =[C2i1, C2i2], mechanisms = [mech_b])
    s2 = CC2.update_species()
    assert X in s2 and Y not in s2 and Z in s2
    assert C2i1 in s2 and C2i2 in s2 and C2 in s2
    assert Complex([X, Y, Z]) in s2
    assert Complex([X, X]) not in s2 and Complex([X, X, Z]) not in s2 
    assert Complex([X, X]) not in s2 and Complex([X, Y]) not in s2 and Complex([X, Z]) not in s2

    #set intermediate states
    CC3 = CombinatorialComplex(final_states = [C2], intermediate_states =[C2i1, C2i2], mechanisms = [mech_b])
    s3 = CC3.update_species()
    assert X in s3 and Y in s3 and Z in s3
    assert C2i1 in s3 and C2i2 in s3 and C2 in s3
    assert Complex([X, Y, Z]) in s3 and Complex([X, X]) in s3 and Complex([X, Y]) in s3
    assert Complex([X, X, Z]) not in s3 and Complex([X, Z]) not in s3

    #set initial and intermediate states
    C4 = Complex([X, Y, Y, Z, Z])
    CC4 = CombinatorialComplex(final_states = [C4], intermediate_states =[Complex([X, Y, Y]), Complex([X, Z, Z])], initial_states = [Complex([X, Y]), Complex([X, Z])], mechanisms = [mech_b])
    s4 = CC4.update_species()
    assert X not in s4 and Y in s4 and Z in s4
    assert Complex([X, Y, Y]) in s4 and Complex([X, Z, Z]) in s4 and Complex([X, Y]) in s4 and Complex([X, Z]) in s4
    assert Complex([Y, Y]) not in s4 and Complex([Z, Z]) not in s4 and Complex([X, Y, Z]) not in s4
    assert Complex([X, Y, Y, Z]) in s4 and Complex([X, Y, Z, Z]) in s4

    #multiple final states
    C5a = Complex([X, X, Y])
    C5b = Complex([X, X, Z])
    CC5 = CombinatorialComplex(final_states = [C5a, C5b], mechanisms = [mech_b])
    s5 = CC5.update_species()
    assert X in s5 and Y in s5 and Z in s5
    assert Complex([X, X]) in s5 and Complex([X, Z]) in s5 and Complex([X, Y]) in s5
    assert Complex([Y, Z]) not in s5


def test_CombinatorialComplex_update_reactions():
    mech_b = One_Step_Binding()
    ku, kb = 1, 1
    params = {"ku":ku, "kb":kb}

    #shortcut to write this faster
    def R(inputs, outputs):
        return Reaction.from_massaction(inputs, outputs, k_forward = kb, k_reverse = ku) 

    #Basic Case
    X, Y, Z = Species("X"), Species("Y"), Species("Z")
    CXYZ = Complex([X, Y, Z])
    CXY = Complex([X, Y])
    CXZ = Complex([X, Z])
    CYZ = Complex([Y, Z])

    CC1 = CombinatorialComplex(final_states = [CXYZ], mechanisms = [mech_b], parameters = params)

    r1 = CC1.update_reactions()
    r1_true = [R([X, Y], [CXY]), R([X, Z], [CXZ]),  R([Y, Z], [CYZ]), R([CXY, Z], [CXYZ]), R([CXZ, Y], [CXYZ]), R([CYZ, X], [CXYZ])]
    assert all([r in r1_true for r in r1]) and all([r in r1 for r in r1_true])

    #Set initial states
    CYZ = Complex([Y, Z])
    CXX = Complex([X, X])
    CXXY = Complex([X, X, Y])
    CXXZ = Complex([X, X, Z])
    CXXYZ = Complex([X, X, Y, Z])
    CC2 = CombinatorialComplex(final_states = [CXXYZ], initial_states =[CXX, CYZ], mechanisms = [mech_b], parameters = params)
    r2 = CC2.update_reactions()
    r2_true = [R([CXX, Y], [CXXY]), R([CXX, Z], [CXXZ]), R([CXXY, Z], [CXXYZ]), R([CXXZ, Y], [CXXYZ]), R([CYZ, X], [CXYZ]), R([CXYZ, X], [CXXYZ])]
    assert all([r in r2_true for r in r2]) and all([r in r2 for r in r2_true])
    
    #set intermediate states
    CC3 = CombinatorialComplex(final_states = [CXXYZ], intermediate_states =[CXX, CYZ], mechanisms = [mech_b], parameters = params)
    r3 = CC3.update_reactions()
    r3_true = [R([X, X], [CXX]), R([Y, Z], [CYZ])] + r2_true
    assert all([r in r3_true for r in r3]) and all([r in r3 for r in r3_true])

    #set initial and intermediate states
    CXXYYZ = Complex([X, X, Y, Y, Z])
    CXXYZZ = Complex([X, X, Y, Z, Z])
    CXXYYZZ = Complex([X, X, Y, Y, Z, Z])
    CC4 = CombinatorialComplex(final_states = [CXXYYZZ], intermediate_states =[CXXYZ], initial_states = [CXX, CYZ], mechanisms = [mech_b], parameters = params)
    r4 = CC4.update_reactions()
    r4_true =  [R([CXX, Y], [CXXY]), R([CXX, Z], [CXXZ]), R([CYZ, X], [CXYZ]), R([CXYZ, X], [CXXYZ]), R([CXXZ, Y], [CXXYZ]), R([CXXY, Z], [CXXYZ]), R([CXXYZ, Y], [CXXYYZ]), R([CXXYZ, Z], [CXXYZZ]), R([CXXYYZ, Z], [CXXYYZZ]), R([CXXYZZ, Y], [CXXYYZZ])]
    assert all([r in r4_true for r in r4]) and all([r in r4 for r in r4_true])

    #multiple final states
    CC5 = CombinatorialComplex(final_states = [CXXY, CXXZ], mechanisms = [mech_b], parameters = params)
    r5 = CC5.update_reactions()
    r5_true = [R([X, X], [CXX]), R([CXX, Y], [CXXY]), R([CXX, Z], [CXXZ]), R([X, Y], [CXY]), R([CXY, X], [CXXY]), R([X, Z], [CXZ]), R([CXZ, X], [CXXZ])]
    assert all([r in r5_true for r in r5]) and all([r in r5 for r in r5_true])











# Copyright (c) 2020, Build-A-Cell. All rights reserved.
# See LICENSE file in the project root directory for details.

from .chemical_reaction_network import ChemicalReactionNetwork
from .components_basic import Protein
from .dna_assembly import DNAassembly
from .global_mechanism import Degredation_mRNA_MM, Dilution
from .mechanism import EmptyMechanism
from .mechanisms_binding import One_Step_Binding
from .mechanisms_enzyme import BasicCatalysis, MichaelisMenten
from .mechanisms_txtl import (OneStepGeneExpression, SimpleTranscription,
                              SimpleTranslation, Transcription_MM,
                              Translation_MM)
from .mixture import Mixture


class ExpressionDilutionMixture(Mixture):
    """A Model for in-vivo Gene Expression without any Machinery (eg Ribosomes, Polymerases, etc.).


    Here transcription and Translation are lumped into one reaction: expression.
    A global mechanism is used to dilute all non-dna species
    """
    def __init__(self, name="", **kwargs):
        """Initializes an ExpressionDilutionMixture instance.

        :param name: name of the mixture
        :param kwargs: keywords passed into the parent Class (Mixture)
        """
        Mixture.__init__(self, name=name, **kwargs)

        # Create default mechanisms for Gene Expression
        dummy_translation = EmptyMechanism(name="dummy_translation", mechanism_type="translation")
        mech_expression = OneStepGeneExpression()
        mech_cat = BasicCatalysis()
        mech_bind = One_Step_Binding()

        default_mechanisms = {
            mech_expression.mechanism_type: mech_expression,
            dummy_translation.mechanism_type: dummy_translation,
            mech_cat.mechanism_type: mech_cat,
            mech_bind.mechanism_type: mech_bind
        }
        self.add_mechanisms(default_mechanisms)

        # Create global mechanism for dilution
        dilution_mechanism = Dilution(name="dilution", filter_dict={"dna": False}, default_on=True)
        global_mechanisms = {"dilution": dilution_mechanism}
        self.add_mechanisms(global_mechanisms)

    def compile_crn(self, **keywords) -> ChemicalReactionNetwork:
        """Overwriting compile_crn to replace transcripts with proteins for all DNA_assemblies.

        Overwriting compile_crn to turn off transcription in all DNAassemblies

        :return: compiled CRN instance
        """
        for component in self.components:
            if isinstance(component, DNAassembly):
                # Only turn off transcription for an Assembly that makes a Protein.
                # Some assemblies might only make RNA!
                if component.protein is not None:
                    # This will turn off transcription and set Promoter.transcript = False
                    # Mechanisms that recieve no transcript but a protein will use the protein instead.
                    component.update_transcript(False)

        # Call the superclass function
        return Mixture.compile_crn(self, **keywords)

class SimpleTxTlDilutionMixture(Mixture):
    """Mixture with continuous dilution for non-DNA species.

    Transcription and Translation are both modeled as catalytic with no cellular machinery.
    mRNA is also degraded via a separate reaction to represent endonucleases
    """
    def __init__(self, name="", **keywords):
        """Initializes a SimpleTxTlDilutionMixture instance.

        :param name: name of the mixture
        :param kwargs: keywords passed into the parent Class (Mixture)
        """
        # Always call the superclass __init__ with **keywords
        Mixture.__init__(self, name=name, **keywords)

        # Create TxTl Mechanisms
        simple_transcription = SimpleTranscription()  # Transcription will not involve machinery
        simple_translation = SimpleTranslation()
        mech_cat = BasicCatalysis()
        mech_bind = One_Step_Binding()
        
        default_mechanisms = {
            simple_transcription.mechanism_type: simple_transcription,
            simple_translation.mechanism_type: simple_translation,
            mech_cat.mechanism_type: mech_cat,
            mech_bind.mechanism_type: mech_bind
        }
        self.add_mechanisms(default_mechanisms)
        
        # Global Dilution Mechanisms
        # By Default Species are diluted S-->0 Unless:
        # They are of type 'dna'
        # They have the attribute 'machinery'
        dilution_mechanism = Dilution(filter_dict={"dna": False}, default_on=True)
        deg_mrna = Dilution(name="rna_degredation", filter_dict={"rna": True}, default_on=False)

        global_mechanisms = {"dilution": dilution_mechanism, "rna_degredation": deg_mrna}
        self.add_mechanisms(global_mechanisms)


class TxTlDilutionMixture(Mixture):
    """A Model for Transcription and Translation with Ribosomes, Polymerases, and Endonucleases labelled as Machinery.

    This model includes a background load "cellular processes" which represents innate loading effects in the cell.
    Effects of loading on cell growth are not modelled.
    Unlike TxTlExtract, has global dilution for non-DNA and non-Machinery
    This model does not include any energy
    """
    def __init__(self, name="", rnap="RNAP", ribosome="Ribo", rnaase="RNAase", **kwargs):
        """Initializes a TxTlDilutionMixture instance.

        :param name: name of the mixture
        :param rnap: name of the RNA polymerase, default: RNAP
        :param ribosome: name of the ribosome, default: Ribo
        :param rnaase: name of the Ribonuclease, default: RNAase
        :param kwargs: keywords passed into the parent Class (Mixture)
        """
        Mixture.__init__(self, name=name, **kwargs)

        # Create Components for TxTl machinery
        self.rnap = Protein(rnap)
        self.ribosome = Protein(ribosome)
        self.rnaase = Protein(rnaase)

        self.rnap.add_attribute("machinery")
        self.ribosome.add_attribute("machinery")
        self.rnaase.add_attribute("machinery")

        # DNAassmbly represents background processes / loading in a cell
        background_parameters = {("transcription", None, "ku"): 50, ("transcription", None, "kb"): 500, ("transcription", None, "ktx"): 0.1,
              ("translation", None, "ku"): 5, ("translation", None, "kb"): 500, ("translation",None, "ktl"): .1,
              ("rna_degredation", None, "ku"): 50, ("rna_degredation", None, "kb"): 500, ("rna_degredation", None, "kdeg"): 0.1}
        BackgroundProcesses = DNAassembly(name="cellular_processes", promoter="average_promoter", rbs="average_rbs", parameters=background_parameters)

        default_components = [
            self.rnap, self.ribosome, self.rnaase, BackgroundProcesses
        ]
        self.add_components(default_components)

        #Create TxTl Mechansisms
        mech_tx = Transcription_MM(rnap = self.rnap.get_species())
        mech_tl = Translation_MM(ribosome = self.ribosome.get_species())
        mech_cat = MichaelisMenten()
        mech_bind = One_Step_Binding()

        #Create Global Dilution Mechanisms
        dilution_mechanism = Dilution(filter_dict = {"dna":False, "machinery":False}, default_on = True)
        mech_rna_deg = Degredation_mRNA_MM(nuclease = self.rnaase.get_species())

        default_mechanisms = {
            mech_tx.mechanism_type: mech_tx,
            mech_tl.mechanism_type: mech_tl,
            mech_cat.mechanism_type: mech_cat,
            mech_bind.mechanism_type:mech_bind,
            mech_rna_deg.mechanism_type:mech_rna_deg,
            "dilution":dilution_mechanism,
        }

        self.add_mechanisms(default_mechanisms)



#  Copyright (c) 2020, Build-A-Cell. All rights reserved.
#  See LICENSE file in the project root directory for details.

from biocrnpyler import CombinatorialComplex, One_Step_Binding, Reaction, CombinatorialConformation
from biocrnpyler import Species, OrderedPolymerSpecies, Complex, PolymerConformation
import pytest


def test_CombinatorialConformation_init():

    #Test getters and setters of properties via init

    X, Y, Z = Species("X"), Species("Y"), Species("Z")
    C = Complex([X, X])
    p1 = OrderedPolymerSpecies([X, Y, Z])
    pc0 = PolymerConformation(polymer = p1)
    pc1 = Complex([pc0.polymers[0][0], pc0.polymers[0][2]]).parent #this gets the PolymerConformation

    CC1 = CombinatorialConformation(initial_states = [pc0], final_states = [pc1])
    
    assert len(CC1.initial_states) == 1 and len(CC1.final_states) == 1

    pc1b = Complex([pc0.polymers[0][2], Z]).parent
    CC1b = CombinatorialConformation(initial_states = [pc0], final_states = [pc1b])
    assert len(CC1.initial_states) == 1 and len(CC1.final_states) == 1

    pc1c = Complex([pc1b.polymers[0][0], pc1b.polymers[0][1]]).parent
    CC1c0 = CombinatorialConformation(initial_states = None, final_states = [pc1b, pc1c])
    assert len(CC1c0.final_states) == 2
    assert len(CC1c0.initial_states) == 1
    CC1c1 = CombinatorialConformation(initial_states = [pc0, pc1b], final_states = [pc1c])
    assert len(CC1c1.initial_states) == 2


    #The following should produce errors:
    p2 = OrderedPolymerSpecies([Z, Y, X])
    pc2 = Complex([pc1.polymers[0][1], PolymerConformation(polymer = p2).polymers[0][1]]).parent
    with pytest.raises(ValueError): #Multiple internal polymers
        CC = CombinatorialConformation(initial_states = [pc1], final_states = [pc2])

    with pytest.raises(ValueError):#Multiple internal polymers
        CC = CombinatorialConformation(initial_states = [pc0, pc1], final_states = [pc2])

    with pytest.raises(ValueError):#Multiple internal polymers
        CC = CombinatorialConformation(initial_states = [pc0], final_states = [pc1], intermediate_states = [pc2])
    with pytest.raises(ValueError):#Multiple internal polymers
        CC = CombinatorialConformation(initial_states = [pc0], final_states = [pc2], excluded_states = [pc1])
    with pytest.raises(ValueError):#Multiple internal polymers
        CC = CombinatorialConformation(initial_states = [pc0], final_states = [pc1], excluded_states = [pc2])

    #Non PolymerConformations passed as initial/final/intermediate states
    with pytest.raises(ValueError):
        CC = CombinatorialConformation(initial_states = [C], final_states = [pc1])

    with pytest.raises(ValueError):
        CC = CombinatorialConformation(initial_states = [pc0], final_states = [C])

    with pytest.raises(ValueError):
        CC = CombinatorialConformation(initial_states = [pc0], final_states = [pc1], intermediate_states = [C])

    with pytest.raises(ValueError):
        CC = CombinatorialConformation(initial_states = [pc0], final_states = [pc1], excluded_states = [C])

    




def pass_test_CombinatorialConformation_compute_complexes_to_add_to_polymer():
    X, Y, Z = Species("X"), Species("Y"), Species("Z")
    p0 = OrderedPolymerSpecies([X])
    p1 = OrderedPolymerSpecies([X, Y, Z])
    p2 = OrderedPolymerSpecies([Z, Y, X])
    p3 = OrderedPolymerSpecies([Complex([X, X]), Y, Complex([X, Z])])

    CC = CombinatorialConformation(initial_states = [], final_states = [])

    #An empty list is returned if the Polymers are the same ("Nothing to add")
    complexes_to_add = CC.compute_complexes_to_add_to_polymer(p0, p0)
    assert len(complexes_to_add) == 0

    #None is returned if the Polymers cannot be converted
    complexes_to_add = CC.compute_complexes_to_add_to_polymer(p0, p1)
    assert complexes_to_add == None
    complexes_to_add = CC.compute_complexes_to_add_to_polymer(p1, p2)
    assert complexes_to_add == None


    #Basic Case
    complexes_to_add = CC.compute_complexes_to_add_to_polymer(p1, p3)
    assert len(complexes_to_add) == 2
    assert (0, [X]) in complexes_to_add and (2, [X]) in complexes_to_add

    #Nested Complex Case
    p4 = OrderedPolymerSpecies([Complex([X, X]), Y, Complex([Z, Complex([X, X])])])
    complexes_to_add = CC.compute_complexes_to_add_to_polymer(p1, p4)
    assert len(complexes_to_add) == 2
    assert (0, [X]) in complexes_to_add and (2, [Complex([X, X])]) in complexes_to_add

    #These cases should give errors
    with pytest.raises(ValueError):
        CC.compute_complexes_to_add_to_polymer(X, p4)
    with pytest.raises(ValueError):
        CC.compute_complexes_to_add_to_polymer(p1, X)

def test_compute_species_changes():
    X, Y, Z, S = Species("X"), Species("Y"), Species("Z"), Species("S")
    C = Complex([X, X])
    p0 = OrderedPolymerSpecies([X, Y, Z, C])
    pc0 = PolymerConformation(polymer = p0)
    c1 = Complex([pc0.polymers[0][0], pc0.polymers[0][1]])
    pc1 = c1.parent
    ind_c1 = pc1.get_polymer_positions(c1, 0)
    c2 = Complex([pc0.polymers[0][0], pc0.polymers[0][1], pc0.polymers[0][2], S, S])
    pc2 = c2.parent
    ind_c2 = pc2.get_polymer_positions(c2, 0)
    c3 = Complex([pc1.polymers[0][2], pc1.polymers[0][3], Z])
    pc3 = c3.parent
    ind_c3 = pc3.get_polymer_positions(c3, 0)
    c4 = Complex([pc0.polymers[0][0], pc0.polymers[0][1], pc0.polymers[0][2], pc0.polymers[0][3], Z])
    pc4 = c4.parent
    ind_c4 = pc4.get_polymer_positions(c4, 0)


    print("pc0", pc0, "\npc1", pc1, "\npc2", pc2, "\npc3", pc3, "\npc4", pc4)


    CC = CombinatorialConformation(initial_states = [pc0], intermediate_states = [pc1], final_states = [pc2, pc3])

    #No additional species added
    SC, MC = CC.compute_species_changes(s0 = pc0, sf = pc1)
    assert (c1, ind_c1) in SC
    assert len(SC[(c1, ind_c1)]) == 0
    assert (c1, ind_c1) in MC
    assert len(MC[(c1, ind_c1)]) == 0

    #External Species added
    SC, MC = CC.compute_species_changes(s0 = pc0, sf = pc2)
    assert (c2, ind_c2) in SC
    assert len(SC[(c2, ind_c2)]) == 2 and S in SC[(c2, ind_c2)]
    assert all([len(MC[cf]) == 0 for cf in MC])

    #Multiple Complexes created
    SC, MC = CC.compute_species_changes(s0 = pc0, sf = pc3)
    assert (pc3.complexes[0], ind_c1) in SC and (pc3.complexes[1], ind_c3) in SC
    assert len(SC[pc3.complexes[0], ind_c1]) == 0
    assert len(SC[pc3.complexes[1], ind_c3]) == 1 and Z in SC[pc3.complexes[1], ind_c3]
    assert all([len(MC[cf]) == 0 for cf in MC])

    #A single complex created, from a conformation with complexes
    SC, MC = CC.compute_species_changes(s0 = pc1, sf = pc3)
    assert (pc3.complexes[0], ind_c1) not in SC and (pc3.complexes[1], ind_c3) in SC
    assert len(SC[pc3.complexes[1], ind_c3]) == 1 and Z in SC[pc3.complexes[1], ind_c3]
    assert len(MC[pc3.complexes[1], ind_c3]) == 0

    #Adding species to an existing Complex
    SC, MC = CC.compute_species_changes(s0 = pc1, sf = pc2)
    assert (pc1.complexes[0], ind_c1) not in SC and (pc2.complexes[0], ind_c2) in SC
    assert len(SC[pc2.complexes[0], ind_c2]) == 2 and S in SC[pc2.complexes[0], ind_c2]
    assert pc1.complexes[0] in MC[pc2.complexes[0], ind_c2]

    #adding species (including inside the polymer) to an existing complex
    SC, MC = CC.compute_species_changes(s0 = pc1, sf = pc4)
    assert (pc1.complexes[0], ind_c1) not in SC and (pc4.complexes[0], ind_c4) in SC
    assert Z in SC[pc4.complexes[0], ind_c4] and len(SC[pc4.complexes[0], ind_c4]) == 1
    assert pc1.complexes[0] in MC[pc4.complexes[0], ind_c4]

    #merging two complexes
    SC, MC = CC.compute_species_changes(s0 = pc3, sf = pc4)
    assert (pc3.complexes[0], ind_c1) not in SC and (pc3.complexes[1], ind_c3) not in SC and (pc4.complexes[0], ind_c4) not in SC
    assert pc3.complexes[0] in MC[pc4.complexes[0], ind_c4] and pc3.complexes[1] in MC[pc4.complexes[0], ind_c4]

    #Cases where compute_species_changes should return False

    #No changes between the conformations
    assert not CC.compute_species_changes(s0 = pc0, sf = pc0)

    #because s0 contains more species/complexes than sf
    assert not CC.compute_species_changes(s0 = pc3, sf = pc0)
    assert not CC.compute_species_changes(s0 = pc3, sf = pc1)
    assert not CC.compute_species_changes(s0 = pc2, sf = pc1)
    assert not CC.compute_species_changes(s0 = pc2, sf = pc3)
    assert not CC.compute_species_changes(s0 = pc3, sf = pc2)


def test_get_combinations_between():
    X, Y, Z, S = Species("X"), Species("Y"), Species("Z"), Species("S")
    C = Complex([X, X])
    p0 = OrderedPolymerSpecies([X, Y, Z, C])
    pc0 = PolymerConformation(polymer = p0)
    c1 = Complex([pc0.polymers[0][0], pc0.polymers[0][1]])
    pc1 = c1.parent
    c2 = Complex([pc0.polymers[0][0], pc0.polymers[0][1], pc0.polymers[0][2], S, S])
    pc2 = c2.parent
    c3 = Complex([pc1.polymers[0][2], pc1.polymers[0][3], Z])
    pc3 = c3.parent
    c4 = Complex([pc0.polymers[0][0], pc0.polymers[0][1], pc0.polymers[0][2], pc0.polymers[0][3], Z])
    pc4 = c4.parent
    c5a = Complex([pc0.polymers[0][0], pc0.polymers[0][2]])
    pc5a = c5a.parent
    c5b = Complex([pc5a.polymers[0][1], pc5a.polymers[0][3]])
    pc5b = c5b.parent


    print("pc0", pc0, "\npc1", pc1, "\npc2", pc2, "\npc3", pc3, "\npc4", pc4)


    #Arguments here don't matter in this test
    CC = CombinatorialConformation(initial_states = [], intermediate_states = [], final_states = [pc3])

    #No additional species added
    perms = CC.get_combinations_between(s0 = pc0, sf = pc1)
    assert len(perms) == 1

    #External Species added
    perms = CC.get_combinations_between(s0 = pc0, sf = pc2)
    assert len(perms) == 1

    #Multiple Complexes created
    perms = CC.get_combinations_between(s0 = pc0, sf = pc3)
    assert len(perms) == 4

    #A single complex created, from a conformation with complexes
    perms = CC.get_combinations_between(s0 = pc1, sf = pc3)
    assert len(perms) == 1

    #Adding species to an existing Complex
    perms = CC.get_combinations_between(s0 = pc1, sf = pc2)
    assert len(perms) == 1

    #adding species (including inside the polymer) to an existing complex
    perms = CC.get_combinations_between(s0 = pc1, sf = pc4)
    assert len(perms) == 1

    #merging two complexes
    perms = CC.get_combinations_between(s0 = pc3, sf = pc4)
    assert len(perms) == 1
   
    #Adding an excluded_state
    #excluded_state matters here
    CC = CombinatorialConformation(initial_states = [], excluded_states = [pc1], final_states = [pc3])
    perms = CC.get_combinations_between(s0 = pc0, sf = pc3)
    assert len(perms) == 2

    #Cases where compute_species_changes should return False
    #s0 contains more species/complexes than sf
    assert len(CC.get_combinations_between(s0 = pc3, sf = pc0)) == 0
    assert len(CC.get_combinations_between(s0 = pc3, sf = pc1)) == 0
    assert len(CC.get_combinations_between(s0 = pc2, sf = pc1)) == 0
    assert len(CC.get_combinations_between(s0 = pc2, sf = pc3)) == 0
    assert len(CC.get_combinations_between(s0 = pc3, sf = pc2)) == 0
    #Partially overlapping sets of monomers are bound
    assert len(CC.get_combinations_between(s0 = pc3, sf = pc5a))== 0
    assert len(CC.get_combinations_between(s0 = pc5a, sf = pc3))== 0
    #These ones are especially tricky because the same Monomers are bound, but in different sets of complexes
    assert len(CC.get_combinations_between(s0 = pc3, sf = pc5b))== 0
    assert len(CC.get_combinations_between(s0 = pc5b, sf = pc3))== 0


def test_update_species_and_reactions():
    params = {"kf":1, "kr":1}

    X, Y, Z, S = Species("X"), Species("Y"), Species("Z"), Species("S")
    C = Complex([X, X])
    p0 = OrderedPolymerSpecies([X, Y, Z, C])
    pc0 = PolymerConformation(polymer = p0)
    c1 = Complex([pc0.polymers[0][0], pc0.polymers[0][1]])
    pc1 = c1.parent
    c2 = Complex([pc0.polymers[0][0], pc0.polymers[0][1], pc0.polymers[0][2], S, S])
    pc2 = c2.parent
    c3 = Complex([pc1.polymers[0][2], pc1.polymers[0][3], Z])
    pc3 = c3.parent
    c4 = Complex([pc0.polymers[0][0], pc0.polymers[0][1], pc0.polymers[0][2], pc0.polymers[0][3], Z])
    pc4 = c4.parent

    #No conformation changes are possible
    CC0 = CombinatorialConformation(initial_states = [], final_states = [pc0], parameters = params)
    species = CC0.update_species()
    reactions = CC0.update_reactions()
    assert len(species) == 0
    assert len(reactions) == 0

    #Only a single binding reaction can occur
    CC1 = CombinatorialConformation(initial_states = [], final_states = [pc1], parameters = params)
    species = CC1.update_species()
    reactions = CC1.update_reactions()
    assert len(species) == 2
    assert pc1 in species and pc0 in species
    assert len(reactions) == 1

    #This should be the same as above, by default
    CC1b = CombinatorialConformation(initial_states = [pc0], final_states = [pc1], parameters = params)
    species = CC1b.update_species()
    reactions = CC1b.update_reactions()
    assert len(species) == 2
    assert pc1 in species and pc0 in species
    assert len(reactions) == 1

    #Test multiple final states
    CC2 = CombinatorialConformation(initial_states = [pc0], final_states = [pc1, pc2], parameters = params)
    species = CC2.update_species()
    reactions = CC2.update_reactions()
    assert len(species) == 4
    assert pc1 in species and pc0 in species and pc2 in species and S in species
    assert len(reactions) == 2

    #Test multiple pathways
    CC3 = CombinatorialConformation(initial_states = [pc0], final_states = [pc3], parameters = params)
    species = CC3.update_species()
    reactions = CC3.update_reactions()
    assert len(species) == 5
    assert pc1 in species and pc0 in species and pc3 in species and Z in species
    assert len(reactions) == 4

    #Test intermediate states
    CC3b = CombinatorialConformation(initial_states = [pc0], intermediate_states = [pc1], final_states = [pc3], parameters = params)
    species = CC3b.update_species()
    reactions = CC3b.update_reactions()
    assert len(species) == 4
    assert pc1 in species and pc0 in species and pc1 in species and Z in species and pc3 in species
    assert len(reactions) == 2

    #Test excluded intermediate states
    CC3c = CombinatorialConformation(initial_states = [pc0], excluded_states = [pc1], final_states = [pc3], parameters = params)
    species = CC3c.update_species()
    reactions = CC3c.update_reactions()
    assert len(species) == 4
    assert pc0 in species and pc1 not in species and Z in species and pc3 in species
    assert len(reactions) == 2

    #Test adding a dead end intermediate species
    CC3c = CombinatorialConformation(initial_states = [pc0], intermediate_states = [pc1, pc2], final_states = [pc3], parameters = params)
    species = CC3c.update_species()
    reactions = CC3c.update_reactions()
    assert len(species) == 6
    assert pc0 in species and pc1 in species and Z in species and pc3 in species and pc2 in species and S in species
    assert len(reactions) == 3

    #Test expanding a pathway through intermediates
    CC4a = CombinatorialConformation(initial_states = [pc0], final_states = [pc4], parameters = params)
    species = CC4a.update_species()
    reactions = CC4a.update_reactions()
    assert len(species) == 3
    assert pc0 in species and pc4 in species
    assert len(reactions) == 1

    CC4b = CombinatorialConformation(initial_states = [pc0], intermediate_states = [pc3], final_states = [pc4], parameters = params)
    species = CC4b.update_species()
    reactions = CC4b.update_reactions()
    assert len(species) == 6
    assert pc0 in species and pc4 in species and pc1 in species and pc3 in species
    assert len(reactions) == 5





"""The classes OrderedPolymer and OrderedMonomer are datastructures used to represent Polymers and their associatd components.

These classes are used by Chemical Reaction Network Species as well as certain Components such as DNA_construct.
"""
import copy
from warnings import warn

class MonomerCollection:
    """
    A class used to represent a collection of OrderedMonomers without any particular structure
    """
    def __init__(self, monomers):
        self.monomers = monomers

    @property
    def monomers(self):
        return self._monomers

    @monomers.setter
    def monomers(self, monomers):
        mon_list = []
        for monomer in monomers:
            assert isinstance(monomer, OrderedMonomer)
            mon_copy = copy.copy(monomer)
            mon_copy.parent = self
            mon_list.append(mon_copy)
        self._monomers = tuple(mon_list)

    
class OrderedPolymer(MonomerCollection):

    """a polymer made up of OrderedMonomers that has a specific order"""
    def __init__(self,parts,default_direction=None):
        """parts can be a list of lists containing 
        [[OrderedMonomer,direction],[OrderedMonomer,direction],...]
        alternatively, you can have a regular list, and the direcitons
        will end up being None"""
        self.default_direction = default_direction
        self.polymer = parts

    @property
    def polymer(self):
        return self._monomers

    @polymer.setter
    def polymer(self, parts):
        polymer = []
        assert(type(parts)==list or type(parts)==tuple), "OrderedPolymer must be instantiated with a list"
        for item in parts:
            if(isinstance(item,list) or isinstance(item,tuple)):
                part = item[0]
                if(len(item)>1):
                    partdir = item[1]
                else:
                    partdir = None
            elif(isinstance(item,OrderedMonomer)):
                part = item
                partdir = item.direction
            else:
                raise ValueError("{} is not an OrderedMonomer or a list of the form [OrderedMonomer,direction]".format(str(item)))
            part_copy = copy.copy(part) #OrderedMonomers are always copied when inserted into an OrderedPolymer
            polymer += [part_copy]
            position = len(polymer)-1
            if(partdir==None):
                partdir = self.default_direction
            part_copy.monomer_insert(self,position,partdir)

        self._monomers = tuple(polymer)


    def __hash__(self):
        hval = 0
        if(not hasattr(self,"_polymer") or len(self._polymer)==0):
            hval = 0
        else:
            hval = sum([a.subhash() for a in self._polymer])
        if(hasattr(self,"name")):
            hval += hash(self.name)

        return hval

    def changed(self):
        #runs whenever anything changed
        pass

    def insert(self,position,part,direction=None):
        part_copy = copy.copy(part) #OrderedMonomers are always copied when inserted into an OrderedPolymer

        if(direction is None):
            direction = part.direction
            
        part_copy.monomer_insert(self,position,direction)
        for subsequent_part in self.polymer[position:]:
            subsequent_part.position += 1
        self.polymer = self.polymer[:position]+(part_copy,)+self.polymer[position:]
        self.changed()

    def replace(self,position,part,direction=None):
        part_copy = copy.copy(part) #OrderedMonomers are always copied when inserted into an OrderedPolymer

        if(direction is None):
            direction = part.direction

        self.polymer[position].remove()
        part_copy.monomer_insert(self,position,direction)
        self.polymer = self.polymer[:position]+(part_copy,)+self.polymer[position+1:]
        self.changed()


    def append(self,part,direction=None):
        part_copy = copy.copy(part) #OrderedMonomers are always copied when inserted into an OrderedPolymer

        if(direction is None):
            if(hasattr(part,"direction")):
                direction = part_copy.direction
            else:
                direction = None
        pos = len(self.polymer)
        self.insert(pos,part_copy,direction)

    def __repr__(self):
        outstr = "polymer("
        for part in self.polymer:
            outstr += str(part)+", direction = "+str(part.direction)+","
        if(outstr[:-1]==","):
            outstr = outstr[:-1]
        outstr += ")"
        return outstr

    def direction_invert(self,dirname):
        if(dirname == "forward"):
            return "reverse"
        elif(dirname == "reverse"):
            return "forward"
        elif(dirname == 0):
            return 1
        elif(dirname == 1):
            return 0
        elif(dirname is None):
            return None
        else:
            warn("didn't know how to invert {}".format(str(dirname)))
            return dirname

    def __len__(self):
        return len(self.polymer)

    def __getitem__(self,ii):
        return self.polymer[ii]

    def __setitem__(self,ii,val):
        self.replace(ii,val,val.direction)

    def __eq__(self,other):
        if(isinstance(other,OrderedPolymer)):
            for item1,item2 in zip(self.polymer,other.polymer):
                if(item1.direction==item2.direction and item1.position==item2.position and type(item1)==type(item2)):
                    pass
                else:
                    return False
            if(len(self.polymer)==len(other.polymer)):
                return True
        return False

    def __contains__(self,item):
        if(item in self.polymer):
            return True
        else:
            return False

    def delpart(self,position):
        part = self.polymer[position]
        part.remove()
        for subsequent_part in self.polymer[position+1:]:
            subsequent_part.position -= 1
        self.polymer = self.polymer[:position] + self.polymer[position+1:]
        self.changed()
        if(hasattr(self,"name") and hasattr(self,"make_name")):
            self.name = self.make_name()

    def reverse(self):
        self.polymer = self.polymer[::-1]
        for ind,part in enumerate(self.polymer):
            part.position = ind
            part.direction = self.direction_invert(part.direction)
        self.changed()

class NamedPolymer(OrderedPolymer):
    """The same as an OrderedPolymer but it has a name"""
    def __init__(self,parts,name,default_direction=None,circular=False):
        self.name = name
        self.circular = circular
        OrderedPolymer.__init__(self=self,parts=parts,default_direction=default_direction)
        
class OrderedMonomer:
    """a unit that belongs to an OrderedPolymer. Each unit has a direction, a location, and a link back to its parent"""
    def __init__(self,direction=None,position=None,parent=None):
        """the default is that the monomer is not part of a polymer"""

        self.parent = None; self.direction = None; self.position = None #Prevents weird testing errors of not having attributes
        self.is_polymer_component = False #by default, we assume that an orderedmonomer is not part of a polymer
        #Set properties correctly
        self.parent = parent
        self.direction = direction
        self.position = position

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, parent):
        if parent is None or isinstance(parent, MonomerCollection):
            self._parent = parent
        else:
            raise ValueError(f"parent must be an MonomerCollection. Recieved {parent}")

    @property
    def direction(self):
        return self._direction
        
    @direction.setter
    def direction(self, direction):
        self._direction = direction

    @property
    def position(self):
        return self._position

    @position.setter
    def position(self, position):
        if self.parent is not None and position is None:
            raise ValueError("{} is part of a polymer with no position!".format(self))
        else:
            self._position = position
    def find_polymer_component(self):
        from .species import ComplexSpecies
        outpolymer = None
        if(isinstance(self,ComplexSpecies)):
            for specie in self.species:
                if(specie.is_polymer_component):
                    if(outpolymer is not None):
                        raise ValueError("multiple species are part of the polymer in the same place!!")
                    else:
                        outpolymer = specie
        if(self.is_polymer_component):
            if(outpolymer is not None):
                raise ValueError("multiple species are part of the polymer in the same place!!")
            else:
                outpolymer = self
        return outpolymer
    def monomer_insert(self,parent:OrderedPolymer,position:int,direction=None):
        if(position is None):
            raise ValueError("{} has no position to be inserted at!".format(self))
        if(direction is None):
            if(self.direction is not None):
                direction  = self.direction
        if(parent is None):
            raise ValueError("{} is trying to be inserted into nothing!".format(self))
        if(self.is_polymer_component is False):
            if(self.find_polymer_component() is None):
                self.is_polymer_component = True
        self.parent = parent
        self.position = position
        self.direction = direction

    def set_dir(self,direction):
        self.direction = direction
        return(self)

    def remove(self):
        self.parent = None
        self.position = None
        self.direction = None
        
        return(self)
    def get_orphan(self):
        """returns a copy of this monomer, except with no parent. But it still has a position and direction"""
        copied_monomer = copy.copy(self)
        copied_monomer.parent = None
        return(copied_monomer)
    def get_removed(self):
        copied_part = copy.copy(self)
        copied_part.parent = None
        copied_part.direction = None
        copied_part.position = None
        if(hasattr(copied_part,"_attributes")):
            copied_part.remove_attribute("forward")
            copied_part.remove_attribute("reverse")
        return copied_part
    def __repr__(self):
        txt = "OrderedMonomer(direction="+str(self.direction)+",position="+\
                                str(self.position)+")"
        return txt

    def __eq__(self,other):
        if(isinstance(other,OrderedMonomer)):
            if(self.direction == other.direction and self.position == other.position and self.parent == other.parent):
                return True
        return False

    def __hash__(self):
        hval = 0
        hval += self.subhash()

        if(self.parent is not None):
            hval+= hash(self.parent)

        return hval

    def subhash(self):
        hval = 0
        hval+= hash(self.position)
        hval += hash(self.direction)
        if(hasattr(self,"name")):
            hval += hash(self.name)
        return hval




#  Copyright (c) 2020, Build-A-Cell. All rights reserved.
#  See LICENSE file in the project root directory for details.

from typing import List, Union

from .component import Component
from .reaction import Reaction
from .species import Complex, Species


class DNA(Component):
    """The DNA class is used to represent a DNA sequence that has a given length.

    Produces no reactions.
    """

    def __init__(self, name, length=0, attributes=None, **keywords):
        """Initialize a DNA object to store DNA related information.

        :param name: Name of the sequence (str)
        :param length: length of the basepairs (int)
        :param attributes: Species attribute
        :param keywords: pass into the parent's (Component) initializer
        """
        self.species = self.set_species(name, material_type="dna", attributes=attributes)
        super().__init__(name=name, length = length, **keywords)

    def get_species(self) -> Species:
        return self.species

    def update_species(self) -> List[Species]:
        species = [self.get_species()]
        return species

    def update_reactions(self) -> List:
        return []

   


class RNA(Component):
    """A class to represent Components made of RNA. Produces no reactions."""

    def __init__(self, name: str, length=0, attributes=None, **keywords):
        """Initialize a RNA object to store RNA related information

        :param name: name of the rna
        :param length: number of basepairs (int)
        :param attributes: Species attribute
        :param keywords: pass into the parent's (Component) initializer
        """

        self.species = self.set_species(name, material_type="rna",
                                        attributes=attributes)
        super().__init__(name=name, length = length, **keywords)

    def get_species(self) -> Species:
        return self.species

    def update_species(self) -> List[Species]:
        species = [self.get_species()]
        return species

    def update_reactions(self) -> List:
        return []

class Protein(Component):
    """A class to represent Components made of Protein. Produces no reactions."""

    def __init__(self, name: str, length=0, attributes=None, **keywords):
        """Initialize a Protein object to store Protein related information.

        :param name: name of the protein
        :param length: length of the protein in number of amino acids
        :param attributes: Species attribute
        :param keywords: pass into the parent's (Component) initializer
        """
        self.species = self.set_species(name, material_type="protein",
                                        attributes=attributes)

        super().__init__(name=name, length = length, **keywords)

    def get_species(self) -> Species:
        return self.species

    def update_species(self) -> List[Species]:
        species = [self.get_species()]
        return species

    def update_reactions(self) -> List:
        return []

class Metabolite(Component):
    def __init__(self, name: str, attributes=None, precursors = None, products = None, **keywords):
        """Initialize a Metabolite object to store Metabolite related information.
        Metabolites look for "production" and "degredation" mechanisms, but will not throw
        an error if none are found.
            
        :param name: name of the protein
        :param attributes: Species attribute
        :param precursors: list of chemical species which are directly transformed into this metabolite via the production mechanism
        :param products: list of chemical species directly produced from this metabolite via the degredation mechanism
        :param keywords: pass into the parent's (Component) initializer
        """

        self.species = self.set_species(name, material_type="metabolite",
                                        attributes=attributes)


        #Set percursor species list
        self.precursors = []
        if precursors is not None:
            for p in precursors:
                if p is None:
                    self.precursors.append(None) #None is a valid precursor representing constuitive production
                else:
                    self.precursors.append(self.set_species(p))

        #Set product species list
        self.products = []
        if products is not None:
            for p in products:
                if p is None:
                    self.products.append(None) #None is a valid product representing total degredation
                else:
                    self.products.append(self.set_species(p))


        Component.__init__(self=self, name=name, **keywords)

    def get_species(self) -> Species:
        return self.species

    def update_species(self) -> List[Species]:
        species = [self.get_species()]
        mech_pathway = self.get_mechanism('metabolic_pathway', optional_mechanism = True)

        if mech_pathway is not None:
            if len(self.precursors) > 0:
                species += mech_pathway.update_species(precursor = self.precursors, product = [self.get_species()], component = self, part_id = self.name+"_production")
            if len(self.products) > 0:
                species += mech_pathway.update_species(precursor = [self.get_species()], product = self.products, component = self, part_id = self.name+"_degredation")

        return species

    def update_reactions(self) -> List:
        reactions = []
        mech_pathway = self.get_mechanism('metabolic_pathway', optional_mechanism = True)

        if mech_pathway is not None:
            if len(self.precursors) > 0:
                reactions += mech_pathway.update_reactions(precursor = self.precursors, product = [self.get_species()], component = self, part_id = self.name+"_production")
            if len(self.products) > 0:
                reactions += mech_pathway.update_reactions(precursor = [self.get_species()], product = self.products, component = self, part_id = self.name+"_degredation")
        return reactions


class ChemicalComplex(Component):
    """A complex forms when two or more species bind together Complexes inherit the attributes of their species."""

    def __init__(self, species: List[Species], name: str=None, material_type="complex",
                 attributes=None, **keywords):
        """Initialize a ChemicalComplex object to store ChemicalComplex related information.

        :param species: list of species inside a complex
        :param name: name of the complex
        :param material_type: option to rename the material_type, default: complex
        :param attributes: Species attribute
        :param keywords: pass into the parent's (Component) initializer
        """
        if not isinstance(species, list) or len(species) < 2:
            raise ValueError(f"Invalid Species {species}. Species must be a list of Species, strings, Component objects.")

        self.internal_species = []  # a list of species inside the complex

        for s in species:
            self.internal_species.append(self.set_species(s))
        if attributes is None:
            attributes = []
        self.species = Complex(species=self.internal_species, name=name, material_type=material_type, attributes=attributes)
        
        if name is None:
            name = self.species.name

        Component.__init__(self=self, name=name, **keywords)

    def get_species(self) -> List[Species]:
        return self.species

    def update_species(self) -> List[Species]:

        mech_b = self.get_mechanism('binding')
        bindee = self.internal_species[0]
        binder = self.internal_species[1:]
        species = mech_b.update_species(binder, bindee, complex_species=self.get_species(), component=self, part_id=self.name)

        return species

    def update_reactions(self) -> List[Reaction]:

        mech_b = self.get_mechanism('binding')
        bindee = self.internal_species[0]
        binder = self.internal_species[1:]
        reactions = mech_b.update_reactions(binder, bindee, complex_species=self.get_species(), component=self, part_id=self.name)
        
        return reactions


class Enzyme(Component):
    """A class to represent Enzymes with multiple substrates and products.

    Assumes the enzyme converts all substrates to a all products at once.
    For example: S1 + S2 + ... + S_N + E --> P1 + P2 + ... + P_M + E.
    For enzymes with multiple enzymatic reactions, create multiple Enzyme Components with the same internal species.
    Uses a mechanism called "catalysis"
    """
    def __init__(self, enzyme: Union[Species, str, Component],
                 substrates: List[Union[Species, str, Component]],
                 products: List[Union[Species, str, Component]], attributes=None, **keywords):
        """Initialize an MultiEnzyme object to store MultiEnzyme related information.

        :param enzyme: name of the enzyme or reference to an Species or Component
        :param substrates: list of (name of the substrate or reference to an Species or Component)
        :param products: list of (name of the product or reference to an Species or Component)
        :param attributes: Species attribute
        :param keywords: pass into the parent's (Component) initializer
        """
        self.enzyme = self.set_species(enzyme, material_type='protein', attributes=attributes)
        self.substrates = substrates
        self.products = products

        Component.__init__(self=self, name=self.enzyme.name, **keywords)

    @property
    def substrates(self) -> List:
        return self._substrates

    @substrates.setter
    def substrates(self, new_substrates: List[Union[Species, str, Component]]):
        if not isinstance(new_substrates, list):
            new_substrates = [new_substrates]
        # convert the new substrates to Species
        self._substrates = [self.set_species(s) for s in new_substrates]

    @property
    def products(self) -> List:
        return self._products

    @products.setter
    def products(self, new_products: List[Union[Species, str, Component]]):
        if not isinstance(new_products, list):
            new_products = [new_products]
        # convert the new products to Products
        self._products = [self.set_species(p) for p in new_products]
    
    def get_species(self) -> Species:
        return self.enzyme

    def update_species(self) -> List[Species]:
        mech_cat = self.get_mechanism('catalysis')

        return mech_cat.update_species(Enzyme=self.enzyme, Sub=self.substrates, Prod=self.products)

    def update_reactions(self) -> List[Reaction]:
        mech_cat = self.get_mechanism('catalysis')
        return mech_cat.update_reactions(Enzyme=self.enzyme, Sub=self.substrates, Prod=self.products,
                                         component=self,  part_id=self.name)




# Copyright (c) 2020, Build-A-Cell. All rights reserved.
# See LICENSE file in the project root directory for details.

from .mechanism import Mechanism
from .reaction import Reaction
from .species import Complex, Species


class One_Step_Reversible_Conformation_Change(Mechanism):
    """A reaction where n binders (A) bind to 1 bindee (B) in one step
       n A + B <--> nA:B
    """
    def __init__(self, name="one_step_conformation_change",
                 mechanism_type="conformation_change"):
        Mechanism.__init__(self, name, mechanism_type)

    def update_species(self, s0, sf, additional_species = None, component = None, part_id = None, **kwords):

        if additional_species is None:
            additional_species = []

        return [s0, sf] + additional_species


    def update_reactions(self, s0, sf, additional_species = None, component = None, part_id = None, **kwords):

        if part_id is None:
            repr(s0)+"-"+repr(sf)

        if additional_species is None:
            additional_species = []

        kf = component.get_parameter("kf", part_id = part_id, mechanism = self)
        kr = component.get_parameter("kr", part_id = part_id, mechanism = self)

        return [Reaction.from_massaction(inputs=[s0]+additional_species, outputs=[sf], k_forward=kf, k_reverse=kr)]


# Copyright (c) 2020, Build-A-Cell. All rights reserved.
# See LICENSE file in the project root directory for details.


from .mechanism import Mechanism
from .reaction import Reaction
from .species import Complex


class BasicCatalysis(Mechanism):
    """Mechanism for the schema S + C --> P + C."""
    def __init__(self, name: str="basic_catalysis", mechanism_type: str="catalysis", **keywords):
        """Initializes a BasicCatalysis instance.

        :param name: name of the Mechanism, default: basic_catalysis
        :param mechanism_type: type of the Mechanism, default: catalysis
        :param keywords:
        """
        # TODO ZAT: remove unused keywords argument
        Mechanism.__init__(self, name, mechanism_type)

    def update_species(self, Enzyme, Sub, Prod = None, **keywords):
        if Prod is None:
            return [Enzyme, Sub]
        else:
            return [Enzyme, Sub, Prod]

    def update_reactions(self, Enzyme, Sub, Prod, component = None, part_id = None, kcat = None, **keywords):
        if part_id is None and component is not None:
            part_id = component.name

        if kcat is None and component is None:
            raise ValueError("Must pass in either a component or kcat.")
        elif kcat is None:
            kcat = component.get_parameter("kcat", part_id = part_id, mechanism = self)

        return [Reaction.from_massaction(inputs=[Enzyme, Sub], outputs=[Enzyme, Prod], k_forward=kcat)]


class BasicProduction(Mechanism):
    """Mechanism for the schema C --> P + C."""
    def __init__(self, name="basic_production", mechanism_type="catalysis", **keywords):
        """Initializes a BasicProduction instance.

        :param name: name of the Mechanism, default: basic_production
        :param mechanism_type: type of the Mechanism, default: catalysis
        :param keywords:
        """
        # TODO ZAT: remove unused keywords argument
        Mechanism.__init__(self, name, mechanism_type)

    def update_species(self, Enzyme, Sub = None, Prod = None, **keywords):
        species = [Enzyme]
        if Prod is not None:
            species += [Prod]
        if Sub is not None:
            species += [Sub]

        return species

    def update_reactions(self, Enzyme, Sub, Prod, component = None, part_id = None, kcat = None, **keywords):
        if part_id is None and component is not None:
            part_id = component.name

        if kcat is None and component is None:
            raise ValueError("Must pass in either a component or kcat.")
        elif kcat is None:
            kcat = component.get_parameter("kcat", part_id = part_id, mechanism = self)

        inputs = [Enzyme]
        outputs = [Enzyme]
        if Prod is not None:
            outputs += [Prod]
        if Sub is not None:
            inputs += [Sub]

        return [Reaction.from_massaction(inputs=inputs, outputs=outputs, k_forward=kcat)]


class MichaelisMenten(Mechanism):
    """Mechanism to automatically generate Michaelis-Menten Type Reactions.

       In the Copy RXN version, the Substrate is not Consumed
       Sub+Enz <--> Sub:Enz --> Enz+Prod
    """

    def __init__(self, name="michalis_menten", mechanism_type="catalysis", **keywords):
        """Initializes a MichaelisMenten instance.

        :param name: name of the Mechanism, default: michalis_menten
        :param mechanism_type: type of the Mechanism, default: catalysis
        :param keywords:
        """
        # TODO ZAT: remove unused keywords argument
        Mechanism.__init__(self, name, mechanism_type)

    def update_species(self, Enzyme, Sub, Prod = None, complex=None, **keywords):
        if complex is None:
            complexS = Complex([Sub, Enzyme])
        else:
            complexS = complex
        if Prod is None:
            return [Enzyme, Sub, complexS]
        else:
            return [Enzyme, Sub, Prod, complexS]

    def update_reactions(self, Enzyme, Sub, Prod, component = None, part_id = None, complex=None, kb=None, ku=None,
                         kcat=None, **keywords):
        # Get Parameters
        if part_id is None and component is not None:
            part_id = component.name

        if component is None and (kb is None or ku is None or kcat is None):
            raise ValueError("Must pass in a Component or values for kb, ku, and kcat.")
        if kb is None:
            kb = component.get_parameter("kb", part_id = part_id, mechanism = self)
        if ku is None:
            ku = component.get_parameter("ku", part_id = part_id, mechanism = self)
        if kcat is None:
            kcat = component.get_parameter("kcat", part_id = part_id, mechanism = self)

        if complex is None:
            complexS = Complex([Sub, Enzyme])
        else:
            complexS = complex

        # Sub + Enz <--> Sub:Enz
        binding_rxn = Reaction.from_massaction(inputs=[Sub, Enzyme],
                                               outputs=[complexS],
                                               k_forward=kb,
                                               k_reverse=ku)
        if Prod is not None:
            # Sub:Enz --> Enz + Prod
            cat_rxn = Reaction.from_massaction(inputs=[complexS],
                                               outputs=[Prod, Enzyme],
                                               k_forward=kcat)
        else:  # Degradation Reaction
            # Sub:Enz --> Enz
            cat_rxn = Reaction.from_massaction(inputs=[complexS],
                                               outputs=[Enzyme],
                                               k_forward=kcat)
        return [binding_rxn, cat_rxn]


class MichaelisMentenReversible(Mechanism):
    """Mechanism to automatically generate Michaelis-Menten Type Reactions with products that can bind to enzymes.

       In the Copy RXN version, the Substrate is not Consumed
       Sub+Enz <--> Sub:Enz <--> Enz:Prod <--> Enz + Prod
    """

    def __init__(self, name="michalis_menten_reverse_binding", mechanism_type="catalysis", **keywords):
        """Initializes a MichaelisMentenReversible instance.

        :param name: name of the Mechanism, default: michalis_menten_reverse_binding
        :param mechanism_type: type of the Mechanism, default: catalysis
        :param keywords:
        """
        # TODO ZAT: remove unused keywords argument
        Mechanism.__init__(self, name, mechanism_type)

    def update_species(self, Enzyme, Sub, Prod, complex=None, complex2 = None, **keywords):
        if complex is None:
            complex1 = Complex([Sub, Enzyme])
        else:
            complex1 = complex
        if complex2 is None:
            complex2 = Complex([Prod, Enzyme])
        else:
            complex2 = complex2
        return [Enzyme, Sub, Prod, complex1, complex2]

    def update_reactions(self, Enzyme, Sub, Prod, component = None, part_id = None, complex=None, complex2 = None, kb=None, ku=None,
                         kcat=None, **keywords):
        #Get Parameters
        if part_id is None and component is not None:
            part_id = component.name

        if component is None and (kb is None or ku is None or kcat is None):
            raise ValueError("Must pass in a Component or values for kb, ku, and kcat.")
        if kb is None:
            kb1 = component.get_parameter("kb1", part_id = part_id, mechanism = self)
            kb2 = component.get_parameter("kb2", part_id = part_id, mechanism = self)
        else:
            kb1, kb2 = kb
        if ku is None:
            ku1 = component.get_parameter("ku1", part_id = part_id, mechanism = self)
            ku2 = component.get_parameter("ku2", part_id = part_id, mechanism = self)
        else:
            ku1, ku2 = ku
        if kcat is None:
            kcat = component.get_parameter("kcat", part_id = part_id, mechanism = self)
            kcat_rev = component.get_parameter("kcat_rev", part_id = part_id, mechanism = self)
        else:
            kcat, kcat_rev = kcat
        

        if complex is None:
            complex1 = Complex([Sub, Enzyme])
        else:
            complex1 = complex
        if complex2 == None:
            complex2 = Complex([Prod, Enzyme])

        # Sub + Enz <--> Sub:Enz
        binding_rxn1 = Reaction.from_massaction(inputs=[Sub, Enzyme],
                                                outputs=[complex1],
                                                k_forward=kb1,
                                                k_reverse=ku1)

        binding_rxn2 = Reaction.from_massaction(inputs=[Prod, Enzyme],
                                                outputs=[complex2],
                                                k_forward=kb2,
                                                k_reverse=ku2)

        # Sub:Enz --> Enz:Prod
        cat_rxn = Reaction.from_massaction(inputs=[complex1],
                                           outputs=[complex2],
                                           k_forward=kcat,
                                           k_reverse=kcat_rev)
        
        return [binding_rxn1, binding_rxn2, cat_rxn]


class MichaelisMentenCopy(Mechanism):
    """In the Copy RXN version, the Substrate is not Consumed.

       Sub+Enz <--> Sub:Enz --> Sub+Enz+Prod
    """
    def __init__(self, name="michalis_menten_copy", mechanism_type="copy", **keywords):
        """Initializes a MichaelisMentenCopy instance.

        :param name: name of the Mechanism, default: michalis_menten_copy
        :param mechanism_type: type of the Mechanism, default: copy
        :param keywords:
        """
        # TODO ZAT: remove unused keywords argument
        Mechanism.__init__(self, name, mechanism_type)

    def update_species(self, Enzyme, Sub, complex=None, Prod = None, **keywords):
        if complex is None:
            complexS = Complex([Sub, Enzyme])
        else:
            complexS = complex
            
        if Prod is None:
            return [Enzyme, Sub, complexS]
        elif(type(Prod)==list):
            return [Enzyme,Sub,complexS]+Prod
        else:
            return [Enzyme, Sub, Prod, complexS]

    def update_reactions(self, Enzyme, Sub, Prod, component = None, part_id = None, complex=None, kb=None, ku=None,
                         kcat=None, **keywords):
        if complex is None:
            complexS = Complex([Sub, Enzyme])
        else:
            complexS = complex

        # Get Parameters
        if part_id is None and component is not None:
            part_id = component.name

        if kb is None and component is not None:
            kb = component.get_parameter("kb", part_id = part_id, mechanism = self)
        if ku is None and component is not None:
            ku = component.get_parameter("ku", part_id = part_id, mechanism = self)
        if kcat is None and component is not None:
            kcat = component.get_parameter("kcat", part_id = part_id, mechanism = self)
        if component is None and (kb is None or ku is None or kcat is None):
            raise ValueError("Must pass in a Component or values for kb, ku, and kcat.")
        # Sub + Enz <--> Sub:Enz
        binding_rxn = Reaction.from_massaction(inputs=[Sub, Enzyme],
                                               outputs=[complexS],
                                               k_forward=kb,
                                               k_reverse=ku)

        # Sub:Enz --> Enz + Prod + Sub
        cat_rxn = Reaction.from_massaction(inputs=[complexS],
                                           outputs=[Sub, Prod, Enzyme],
                                           k_forward=kcat)

        return [binding_rxn, cat_rxn]




# Copyright (c) 2020, Build-A-Cell. All rights reserved.
# See LICENSE file in the project root directory for details.

import itertools as it

from .mechanism import Mechanism
from .reaction import Reaction
from .species import Complex, Species, WeightedSpecies


class One_Step_Cooperative_Binding(Mechanism):
    """A reaction where n binders (A) bind to 1 bindee (B) in one step
       n A + B <--> nA:B
    """
    def __init__(self, name="one_step_cooperative_binding",
                 mechanism_type="cooperative_binding"):
        Mechanism.__init__(self, name, mechanism_type)

    def update_species(self, binder, bindee, complex_species = None, cooperativity=None, component = None, part_id = None, **kwords):

        if part_id is None:
            part_id = repr(binder)+"-"+repr(bindee)

        if cooperativity is None and component != None:
            cooperativity = component.get_parameter("cooperativity", part_id = part_id, mechanism = self, return_numerical = True)
        elif component is None and cooperativity is None:
            raise ValueError("Must pass in a Component or values for cooperativity")

        complexS = None
        if isinstance(complex_species, Species):
            complexS = complex_species
        elif complex_species is not None:
            raise TypeError("complex_species keyword must be a Species, or None.")

        if complexS is None:
            complexS = Complex([binder]*int(cooperativity)+[bindee])

        
        return [binder, bindee, complexS]

    def update_reactions(self, binder, bindee, complex_species = None, component = None, kb = None, ku = None, part_id = None, cooperativity=None, **kwords):

        # Get Parameters
        if part_id is None:
            part_id = repr(binder)+"-"+repr(bindee)
        if kb is None and component is not None:
            kb = component.get_parameter("kb", part_id = part_id, mechanism = self)
        if ku is None and component is not None:
            ku = component.get_parameter("ku", part_id = part_id, mechanism = self)
        if cooperativity is None and component is not None:
            cooperativity = component.get_parameter("cooperativity", part_id = part_id, mechanism = self, return_numerical = True)
        if component is None and (kb is None or ku is None or cooperativity is None):
            raise ValueError("Must pass in a Component or values for kb, ku, and coopertivity.")

        complexS = None
        if complex_species is None:
            complex_name = None
        elif isinstance(complex_species, Species):
            complexS = complex_species
        else:
            raise TypeError("complex_species keyword must be a str, Species, or None.")

        if complexS is None:
            complexS = Complex([binder]*int(cooperativity)+[bindee], name = complex_name)

        inputs = [WeightedSpecies(species=binder, stoichiometry=cooperativity),
                  WeightedSpecies(species=bindee, stoichiometry=1)]

        rxns = [Reaction.from_massaction(inputs=inputs, outputs=[complexS], k_forward=kb, k_reverse=ku)]
        return rxns


class Two_Step_Cooperative_Binding(Mechanism):
    """A reaction where n binders (s1) bind to 1 bindee (s2) in two steps.

       n A <--> nx_A
       nx_A <--> nx_A:B
    """
    def __init__(self, name="two_step_cooperative_binding",
                 mechanism_type="cooperative_binding"):
        Mechanism.__init__(self, name, mechanism_type)

    def update_species(self, binder, bindee, component = None, complex_species = None, n_mer_species = None, cooperativity=None, part_id = None, **keywords):

        if part_id is None:
            part_id = repr(binder)+"-"+repr(bindee)

        if cooperativity is None and component != None:
            cooperativity = component.get_parameter("cooperativity", part_id = part_id, mechanism = self, return_numerical = True)
        elif component is None and cooperativity is None:
            raise ValueError("Must pass in a Component or values for cooperativity")

        n_mer = None
        if isinstance(n_mer_species, Species):
            n_mer = n_mer_species
        elif n_mer_species is not None:
            raise TypeError("n_mer_species keyword nust be a Species, or None. Not "+str(n_mer_species))

        if n_mer is None:
            n_mer = Complex(cooperativity*[binder])

        complexS = None
        if isinstance(complex_species, Species):
            complexS = complex_species
        elif not complex_species is None:
            raise TypeError("complex_species keyword must be a Species, or None. Not "+str(complex_species))

        if complexS is None:
            complexS = Complex([n_mer, bindee])
        return [binder, bindee, complexS, n_mer]

    def update_reactions(self, binder, bindee, kb = None, ku = None, component = None, part_id = None, \
                                cooperativity=None, complex_species = None, n_mer_species = None, **keywords):
        """
        Returns reactions:
        cooperativity binder <--> n_mer, kf = kb1, kr = ku1
        n_mer + bindee <--> complex, kf = kb2, kr = ku2
        :param s1:
        :param s2:
        :param kb:
        :param ku:
        :param cooperativity:
        :param keywords:
        :return:
        """

        if part_id is None:
            repr(binder)+"-"+repr(bindee)
        if (kb is None or ku is None or cooperativity is None) and component != None:
            kb1 = component.get_parameter("kb1", part_id = part_id, mechanism = self)
            kb2 = component.get_parameter("kb2", part_id = part_id, mechanism = self)
            ku1 = component.get_parameter("ku1", part_id = part_id, mechanism = self)
            ku2 = component.get_parameter("ku2", part_id = part_id, mechanism = self)
            cooperativity = component.get_parameter("cooperativity", part_id = part_id, mechanism = self, return_numerical = True)
        elif component is None and (kb is None or ku is None or cooperativity is None):
            raise ValueError("Must pass in a Component or values for kb, ku, and cooperativity")
        elif len(kb) != len(ku) != 2:
            raise ValueError("kb and ku must contain 2 values each for "
                             "two-step binding")
        else:
            kb1, kb2 = kb
            ku1, ku2 = ku

        binder, bindee, complexS, n_mer = self.update_species(binder, bindee, component = component, \
                                                    complex_species = complex_species, n_mer_species = n_mer_species,\
                                                    cooperativity=cooperativity, part_id = part_id, **keywords)

        inputs_for_rxn1 = [WeightedSpecies(species=binder, stoichiometry=cooperativity)]
        rxns = [
            Reaction.from_massaction(inputs=inputs_for_rxn1, outputs=[n_mer], k_forward=kb1, k_reverse=ku1),
            Reaction.from_massaction(inputs=[n_mer, bindee], outputs=[complexS], k_forward=kb2, k_reverse=ku2)
        ]

        return rxns

class Combinatorial_Cooperative_Binding(Mechanism):
    """a reaction where some number of binders bind combinatorially to a bindee"""
    def __init__(self,name="Combinatorial_Cooperative_binding",
                            mechanism_type="cooperative_binding"):
        """Initializes a Combinatorial_Cooperative_Binding instance

        :param name: name of the Mechanism, default: Combinatorial_Cooperative_binding
        :param mechanism_type: type of the Mechanism, default: cooperative_binding
        """
        Mechanism.__init__(self,name,mechanism_type)

    def make_cooperative_complex(self,combo,bindee,cooperativity):
        """given a list of binders and their cooperativities, make a complex
        that contains the binders present N number of times where N is
        each one's cooperativity."""
        complexed_species_list = []
        for binder in combo:
            binder_cooperativity = int(cooperativity[binder.name])
            #I hope that cooperativity is an int! what if it isn't
            if(binder_cooperativity > 1):
                complexed_species_list += [binder]*binder_cooperativity
            else:
                complexed_species_list += [binder]
        complexed_species_list += [bindee]
        if(len(complexed_species_list)==1):
            myspecies = complexed_species_list[0]
        else:
            myspecies = Complex(complexed_species_list)
        return myspecies

    def update_species(self,binders,bindee,cooperativity=None,\
                            component = None, part_id = None, **kwords):
        cooperativity_dict = {}
        out_species = []
        for binder in binders:
            binder_partid = part_id+"_"+binder.name
            if ((cooperativity is None) or (type(cooperativity)==dict and binder_partid not in cooperativity) \
                                                                                        and (component is not None)):
                #here we are extracting the relevant cooperativity value from the dictionary which should be passed
                #in as the cooperativity argument
                coop_val = component.get_parameter("cooperativity", part_id = binder_partid, mechanism = self, return_numerical = True)
            elif type(cooperativity)==dict and binder_partid in cooperativity:
                coop_val = cooperativity[binder_partid]
            if component is None and ( cooperativity is None):
                raise ValueError("Must pass in a Component or values for kb, ku, and coopertivity.")
            cooperativity_dict[binder.name]=coop_val
        
        for i in range(1, len(binders)+1):
            for combo in it.combinations(binders,i):
                #go through every possible combination of reactants and dna and make
                #all the complexes
                out_species += [self.make_cooperative_complex(combo,bindee,cooperativity_dict)]
        return out_species
    
    def update_reactions(self,binders,bindee,component=None,kbs=None,kus=None,\
                                                            part_id = None,cooperativity=None,**kwords):
        binder_params = {}
        for binder in binders:
            binder_partid = part_id+"_"+binder.name
            if ((type(kbs)==dict and binder not in kbs) or (type(kbs)!=dict and component is not None)):
                kb = component.get_parameter("kb", part_id = binder_partid, mechanism = self)
            elif(type(kbs)==dict and binder in kbs):
                kb = kbs[binder.name]
            elif(type(kbs)!=dict and component is None):
                raise ValueError("Must pass in a Component or values for kb, ku, and coopertivity.")
            if ((type(kus)==dict and binder not in kus) or (kus is None and component is not None)):
                ku = component.get_parameter("ku", part_id = binder_partid, mechanism = self)
            elif(type(kus)==dict and binder in kus):
                ku = kus[binder.name]
            elif(type(kus)!=dict and component is None):
                raise ValueError("Must pass in a Component or values for kb, ku, and coopertivity.")
            if ((cooperativity is None) or (type(cooperativity)==dict and binder.name not in cooperativity)  \
                                                                                        and component is not None):
                coop_val = component.get_parameter("cooperativity", part_id = binder_partid, mechanism = self, return_numerical = True)
            elif type(cooperativity)==dict and binder.name in cooperativity:
                coop_val = cooperativity[binder.name]
            if component is None and (kb is None or ku is None or cooperativity is None):
                raise ValueError("Must pass in a Component or values for kb, ku, and coopertivity.")
            binder_params[binder] = {"kb":kb,"ku":ku,"cooperativity":coop_val}
        #out_rxns = []
        rxndict = {}
        coop_dict = {a.name:binder_params[a]["cooperativity"] for a in binder_params}
        for i in range(1, len(binders)+1):
            for combo in it.combinations(binders,i):
                #come up with all combinations of binders
                
                product = self.make_cooperative_complex(combo,bindee,coop_dict)
                #this is the complex which becomes the product
                for binder in combo:
                    #then in each case, one binder can be added to make
                    #this combination
                    reactant = tuple(set(combo)-set([binder]))
                    rxn_prototype = (binder,reactant)
                    
                    #this part makes a describer of the reaction; which reactants are combining?
                    #print(rxn_prototype)
                    #print(rxndict)
                    if(rxn_prototype in rxndict):
                        #if we already did this reaction then forget about it
                        continue
                    else:
                        reactant_complex = self.make_cooperative_complex(reactant,bindee,coop_dict)

                        inputs = [WeightedSpecies(species=binder, stoichiometry=binder_params[binder]["cooperativity"]),
                                  WeightedSpecies(species=reactant_complex, stoichiometry=1)]

                        reaction = Reaction.from_massaction(inputs=inputs, outputs=[product],
                                                            k_forward=binder_params[binder]["kb"],
                                                            k_reverse=binder_params[binder]["ku"])
                        rxndict[rxn_prototype]=reaction
        return [rxndict[a] for a in rxndict]

        
class One_Step_Binding(Mechanism):
    """
    A mechanism to model the binding of a list of species, eg
    S1 + S2 ... SN <--> S1:S2:...:SN
    """
    def __init__(self, name="one_step_binding",
                 mechanism_type="binding"):
        Mechanism.__init__(self, name, mechanism_type)

    def update_species(self, binder,bindee, component = None, complex_species = None, part_id = None, **keywords):
        if(not isinstance(binder,list)):
            binder = [binder]
        if(not isinstance(bindee,list)):
            bindee = [bindee]
        species = binder+bindee
        if part_id is None:
            part_id = ""
            for s in species:
                part_id += s.name+"_"
            part_id = part_id[:-1]

        if complex_species is None:
            complex_species = Complex(species)

        return species + [complex_species]

    def update_reactions(self, binder,bindee, component = None, complex_species = None, part_id = None, kb = None, ku = None, **keywords):
        if(not isinstance(binder,list)):
            binder = [binder]
        if(not isinstance(bindee,list)):
            bindee = [bindee]
        species = binder+bindee
        if part_id is None:
            part_id = ""
            for s in species:
                part_id += s.name+"_"
            part_id = part_id[:-1]

        if (kb is None or ku is None) and component != None:
            kb = component.get_parameter("kb", part_id = part_id, mechanism = self)
            ku = component.get_parameter("ku", part_id = part_id, mechanism = self)
        elif component is None and (kb is None or ku is None):
            raise ValueError("Must pass in a Component or values for kb and ku")

        if complex_species is None:
            complex_species = Complex(species)

        return [Reaction.from_massaction(inputs=species, outputs=[complex_species], k_forward=kb, k_reverse=ku)]




# Copyright (c) 2020, Build-A-Cell. All rights reserved.
# See LICENSE file in the project root directory for details.

from .chemical_reaction_network import Species
from .component import Component
from .dna_part import DNA_part


class CDS(DNA_part):
    def __init__(self,name,protein=None,no_stop_codons=None, **keywords):
        """a CDS is a sequence of DNA that codes for a protein"""
        self.name = name
        DNA_part.__init__(self,name,no_stop_codons=no_stop_codons, **keywords)
        #TODO use set_species()
        if(protein is None):
            self.protein = Species(name,material_type="protein")
        elif(isinstance(protein,str)):
            self.protein = Species(protein,material_type="protein")
        elif(isinstance(protein,Component)):
            self.protein=protein.get_species()
    def update_species(self):
        return [self.protein]
    def update_reactions(self):
        return []
    def get_species(self):
        return self.protein



# __init__.py - initialization of biocrnpyler toolbox
# RMM, 11 Aug 2018

from .chemical_reaction_network import *
from .component import *
# Core components
from .components_basic import *
from .dna_assembly import *
from .dna_construct import *
from .dna_part import *
from .dna_part_cds import *
from .dna_part_misc import *
from .dna_part_promoter import *
from .dna_part_rbs import *
from .dna_part_terminator import *
from .construct_explorer import *
from .integrase_enumerator import *
from .components_combinatorial_complex import *
from .components_combinatorial_conformation import *

from .global_mechanism import *
from .mechanism import *
#core mechanisms
from .mechanisms_binding import *
from .mechanisms_enzyme import *
from .mechanisms_txtl import *
from .mechanisms_integrase import *
# Core classes
from .mixture import *
from .mixtures_cell import *
from .mixtures_extract import *
from .parameter import *
from .plotting import *
from .polymer import *
from .propensities import *
from .reaction import *

from .sbmlutil import *
from .species import *
from .compartments import *
from .utils import *

#checking for nonexistant plotting-related modules now happens in plotting.py
from .component_enumerator import *


# plotting.py - plotting utilities
# ASS, 21 Apr 2020
#
# This file contains some utility functions for plotting CRNs
#
# Copyright (c) 2018, Build-A-Cell. All rights reserved.
# See LICENSE file in the project root directory for details.

import math
import random
import statistics
from warnings import warn

from .components_basic import Protein, DNA, RNA
from .dna_part_cds import CDS
from .dna_part_misc import IntegraseSite, Origin, Operator, UserDefined
from .dna_part_promoter import Promoter
from .dna_part_rbs import RBS
from .dna_part_terminator import Terminator
from .propensities import MassAction
from .species import ComplexSpecies, Species
from .polymer import OrderedPolymer
from .dna_construct import Construct
from .utils import member_dictionary_search
import io
import base64

import copy
HAVE_MATPLOTLIB = False
try:
    import matplotlib.pyplot as plt
    from matplotlib import cm

    HAVE_MATPLOTLIB = True
except ModuleNotFoundError:
    pass
PLOT_DNA = False
try:
    import dnaplotlib as dpl
    PLOT_DNA = True
except ModuleNotFoundError:
    pass

if(PLOT_DNA and not HAVE_MATPLOTLIB):
    PLOT_DNA = False

PLOT_NETWORK = False
try:
    import networkx as nx
    from bokeh.models import (BoxSelectTool, Circle, EdgesAndLinkedNodes,
                              HoverTool, MultiLine, NodesAndLinkedEdges,
                              PanTool, Plot, Range1d, TapTool,
                              WheelZoomTool)
    from bokeh.models.markers import Square
    from bokeh.plotting import from_networkx
    from bokeh.palettes import Spectral4
    from bokeh.io import export_svgs, output_notebook
    from fa2 import ForceAtlas2
    PLOT_NETWORK = True
except ModuleNotFoundError:
    pass

empty_set_base64 = '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'

def updateLimits(limits, xvalues):
    for value in xvalues:
        if(value < limits[0]):
            limits[0] = value
        if(value > limits[1]):
            limits[1] = value
    return limits


def makeArrows2(graph_renderer, graph, positions, headsize=3, headangle=math.pi/6, make_arrows = True):
    """this function draws an arrow shape at the end of graph lines"""
    xs, ys = [], []
    xbounds = [0, 0]
    ybounds = [0, 0]
    for edge in graph.edges:
        # iterate through all the edges
        from_node = edge[0]
        to_node = edge[1]
        from_x = positions[from_node][0]
        from_y = positions[from_node][1]
        to_x = positions[to_node][0]
        to_y = positions[to_node][1]
        updateLimits(xbounds, [from_x, to_x])
        updateLimits(ybounds, [from_y, to_y])
        # above, we get all the variables
        # below, we are assuming the "to" position is the middle of the
        # coordinate space
        ydif = from_y-to_y
        xdif = from_x-to_x
        # next we calculate the angle from the destination node
        # to the source node
        angl = math.atan2(ydif, xdif)
        # the arrow consists of three added points, one on either side
        # of the line and one in the middle
        p1x = to_x+headsize*math.cos(angl+headangle)  # left side of the arrow
        p1y = to_y+headsize*math.sin(angl+headangle)  # left side of the arrow
        p2x = to_x+headsize*math.cos(angl-headangle)  # right side of the arrow
        p2y = to_y+headsize*math.sin(angl-headangle)  # right side of the arrow
        p3x = to_x+headsize*.7*math.cos(angl)  # middle of the arrow
        p3y = to_y+headsize*.7*math.sin(angl)  # middle of the arrow
        # 'xs' is a list of lists which represent each line from node to node
        xs.append([from_x, p3x, p1x, to_x, p2x, p3x])
        # 'ys' is the same thing except the y positions
        ys.append([from_y, p3y, p1y, to_y, p2y, p3y])
    # this part replaces the lines with the ones made by this function
    if make_arrows:
        graph_renderer.edge_renderer.data_source.data['xs'] = xs
        graph_renderer.edge_renderer.data_source.data['ys'] = ys
    return xbounds, ybounds


def graphPlot(DG,DGspecies,DGreactions,plot,layout="force",positions=None,plot_species = True, plot_reactions = True, plot_edges = True, plot_arrows = True,\
              species_glyph_size = 12, reaction_glyph_size = 8, posscale = 1.0,layoutfunc=None,iterations=2000,rseed=30,show_species_images=False):
    """given a directed graph, plot it!
    Inputs:
    DG: a directed graph of type DiGraph
    DGspecies: a directed graph which only contains the species nodes
    DGreactions: a directed graph which only contains the reaction nodes
    plot: a bokeh plot object
    layout: graph layout function. 
                'force' uses fa2 to push nodes apart
                'circle' plots the nodes and reactions in two overlapping circles, with the reactions on the inside of the circle
                'custom' allows user input "layoutfunc". Internally, layoutfunc is passed the three inputs (DG, DGspecies, DGreactions)
                                                        and should output a position dictionary with node {<node number>:(x,y)}

    positions: a dictionary of node names and x,y positions. this gets passed into the layout function
    posscale: multiply the scaling of the plot. This only affects the arrows because the arrows are a hack :("""
    random.seed(rseed)
    if(not PLOT_NETWORK):
        warn("network plotting disabled because some libraries are not found")
        return
    if(layout == "force"):
        # below are parameters for the force directed graph visualization
        forceatlas2 = ForceAtlas2(
            # Behavior alternatives
            outboundAttractionDistribution=True,  # Dissuade hubs
            linLogMode=False,  # NOT IMPLEMENTED
            # Prevent overlap (NOT IMPLEMENTED)
            adjustSizes=False,
            edgeWeightInfluence=1.0,

            # Performance
            jitterTolerance=1.0,  # Tolerance
            barnesHutOptimize=True,
            barnesHutTheta=1.2,
            multiThreaded=False,  # NOT IMPLEMENTED

            # Tuning
            scalingRatio=2.4*posscale,
            strongGravityMode=False,
            gravity=1.0,

            # Log
            verbose=False)

        positions = forceatlas2.forceatlas2_networkx_layout(
            DG, pos=positions, iterations=iterations)
    elif(layout == "circle"):
        positions = nx.circular_layout(DGspecies, scale=50*posscale)
        positions.update(nx.circular_layout(DGreactions, scale=35*posscale))
    elif(layout == "custom"):
        positions = layoutfunc(DG, DGspecies, DGreactions)
    reaction_renderer = from_networkx(DGreactions, positions, center=(0, 0))
    species_renderer = from_networkx(DGspecies, positions, center=(0, 0))
    edges_renderer = from_networkx(DG, positions, center=(0, 0))

    #Set xbounds and ybounds:
    xbounds = [0, 0]
    ybounds = [0, 0]
    for n in positions:
        xbounds[0] = min([xbounds[0], positions[n][0]])
        xbounds[1] = max([xbounds[1], positions[n][0]])
        ybounds[0] = min([ybounds[0], positions[n][1]])
        ybounds[1] = max([ybounds[1], positions[n][1]])

    max_glyph = max([reaction_glyph_size, species_glyph_size])
    xbounds[0] -= max_glyph
    xbounds[1] += max_glyph
    ybounds[0] -= max_glyph
    ybounds[1] += max_glyph

    # edges
    edges_renderer.node_renderer.glyph = Circle(
        size=species_glyph_size, line_alpha=0, fill_alpha=0, fill_color="color")
    edges_renderer.edge_renderer.glyph = MultiLine(
        line_alpha=0.2, line_width=4, line_join="round", line_color="color")
    edges_renderer.edge_renderer.selection_glyph = MultiLine(
        line_color=Spectral4[2], line_width=5, line_join="round")
    edges_renderer.edge_renderer.hover_glyph = MultiLine(
        line_color=Spectral4[1], line_width=5, line_join="round")
    if plot_arrows:
        xbounds_a, ybounds_a = makeArrows2(
            edges_renderer, DG, positions, headsize=5)  # make the arrows!
        xbounds[0] = min([xbounds[0], xbounds_a[0]])
        xbounds[1] = max([xbounds[1], xbounds_a[1]])
        ybounds[0] = min([ybounds[0], ybounds_a[0]])
        ybounds[1] = max([ybounds[1], ybounds_a[1]])

    # we want to find the middle of the graph and plot a square that is 1:1 aspect ratio

    # find the midpoint of the graph
    xmid = statistics.mean(xbounds)
    ymid = statistics.mean(ybounds)
    # now, subtract the middle from the edges
    xmiddlized = [a-xmid for a in xbounds]
    ymiddlized = [a-ymid for a in ybounds]
    # now, find the biggest dimension
    absdim = max([abs(a) for a in xmiddlized+ymiddlized])
    xlim = [xmid-absdim*1.05, xmid + absdim*1.05]
    ylim = [ymid-absdim*1.05, ymid + absdim*1.05]
    # now set it on the plot!
    plot.x_range = Range1d(xlim[0], xlim[1])
    plot.y_range = Range1d(ylim[0], ylim[1])

    # reactions
    reaction_renderer.node_renderer.glyph = Square(
        size=reaction_glyph_size, fill_color="color")
    reaction_renderer.node_renderer.selection_glyph = Square(
        size=reaction_glyph_size, fill_color=Spectral4[2])
    reaction_renderer.node_renderer.hover_glyph = Square(
        size=reaction_glyph_size, fill_color=Spectral4[1])

    # nodes
    species_renderer.node_renderer.glyph = Circle(size=12, fill_color="color")
    species_renderer.node_renderer.selection_glyph = Circle(size=15, fill_color=Spectral4[2])
    species_renderer.node_renderer.hover_glyph = Circle(size=15, fill_color=Spectral4[1])
    
    #this part adds the interactive elements that make it so that the lines are highlighted 
    #when you mouse over and click
    edge_hover_tool = HoverTool(tooltips= None,renderers=[edges_renderer])
    if( not show_species_images):
        species_hover_tool = HoverTool(tooltips=[("name", "@species"), ("type", "@type")],\
                                        renderers=[species_renderer],attachment="right")
    else:
        species_hover_tool = HoverTool(tooltips='<div><div> <img src="data:image/png;base64,@image" style="float: left; margin: 0px 0px 0px 0px;"></img></div></div>',\
                                        renderers=[species_renderer],attachment="right")
    rxn_hover_tool = HoverTool(tooltips=[("reaction", "@species"), ("type", "@type"),("k_f","@k"),("k_r","@k_r")],\
                                        renderers=[reaction_renderer],attachment="right")
    
    plot.add_tools(edge_hover_tool,species_hover_tool,rxn_hover_tool, TapTool(), BoxSelectTool(),PanTool(),WheelZoomTool())

    edges_renderer.selection_policy = NodesAndLinkedEdges()
    edges_renderer.inspection_policy = EdgesAndLinkedNodes()

    if plot_edges:
        plot.renderers.append(edges_renderer)
    if plot_reactions:
        plot.renderers.append(reaction_renderer)
    if plot_species:
        plot.renderers.append(species_renderer)

def generate_networkx_graph(CRN,useweights=False,use_pretty_print=False,pp_show_material=True,
                            pp_show_rates=True,pp_show_attributes=True, pp_show_compartments=True,
                            colordict=None,reactioncolordict = None, imagedict = None):
    """generates a networkx DiGraph object that represents the CRN.
    input:
    ==========================
    CRN: a CRN from mixture.get_model() for example
    useweights: this will attempt to represent the reaction rates by the length of edges.
                short edges are fast rates. It doesn't look very good usually
    use_pretty_print: this uses the "pretty print" function to represent reactions and nodes a bit cleaner
    the next three parameters are pretty_print parameters
        pp_show_material: default false because this is listed in "type"
        pp_show_rates: default true because this is useful information
        pp_show_attributes
    colordict: a dictionary containing which node types are what color based upon the following keywords:
        Keywords are chosen to match species.material_type
            {"complex": "cyan",
            "protein": "green",
            "dna": "grey",
            "rna": "orange",
            "ligand": "pink",
            "phosphate": "yellow",
            "nothing":"purple"}

    When using a custom colordict, the following attributes will be checked to find colors with the first keys taking precedence:
        repr(species): "color"
        species.name: "color"
        (species.material_type, tuple(species.attributes)): "color"
        species.material_type: "color"
        tuple(species.attributes): "color"

    imagedict is a dictionary which contains species and their corresponding image representations.
    This is the output generated by CRNPlotter.renderMixture()

    output:
    ==================
    CRNgraph: the DiGraph object containing all nodes and edges
    CRNspeciesonly: a DiGraph object with only species
    CRNreactionsonly: a DiGraph object with only reactions
    """
    if(not PLOT_NETWORK):
        warn("network plotting disabled because some libraries are not found")
        return None, None, None
    if (colordict is None):
        colordict = {"complex": "cyan", "protein": "green",
                     "dna": "grey", "rna": "orange",
                     "ligand": "pink", "phosphate": "yellow", "nothing": "purple"}
    CRNgraph = nx.DiGraph()
    allnodenum = 1  # every node has an index
    alledgenum = 0 #every edge has an index
    # this starts at 1 because "nothing" is node 0
    nodedict = {}  # this is so that we can write out the reactions in
    # the reaction "species" field
    # it has {species:index}
    rxnlist = []  # list of numbers corresponding to only reaction nodes
    # sometimes reactions have no products. I want this to be represented in the graph with this
    # "nothing" node. However, usually we are making degradation reactions which yield the
    # degradation enzyme, so then it doesn't go to nothing. This means actually this node
    # isn't use for anything. But i think it's good to have just in case.
    
    default_species_color = "grey"
    if (reactioncolordict is None or 'reaction' not in reactioncolordict) and 'reaction' not in colordict:
        default_reaction_color = "cornflowerblue"
    elif 'reaction' in colordict:
        default_reaction_color = colordict['reaction']
    else:
        default_reaction_color = reactioncolordict['reaction']

    if (reactioncolordict is None or 'edge' not in reactioncolordict) and 'edge' not in colordict:
        default_edge_color = 'gray'
    elif 'edge' in colordict:
        default_edge_color = colordict['edge']
    else:
        default_edge_color = reactioncolordict['edge']

    nodedict["nothing"] = 0
    CRNgraph.add_node(0)
    CRNgraph.nodes[0]["type"] = "nothing"
    CRNgraph.nodes[0]["species"] = "nothing"
    CRNgraph.nodes[0]["image"] = empty_set_base64
    if("nothing" in colordict):
        CRNgraph.nodes[0]["color"] = colordict["nothing"]
    for species in CRN.species:
        # add all species first
        species_color = member_dictionary_search(species,colordict)
        if(species_color is None):
            species_color = default_species_color

        nodedict[species] = allnodenum
        CRNgraph.add_node(allnodenum)
        CRNgraph.nodes[allnodenum]["type"]=str(species.material_type)
        if((imagedict is not None) and (species in imagedict)):
            CRNgraph.nodes[allnodenum]["image"]= imagedict[species].decode()
        if(not use_pretty_print):
            CRNgraph.nodes[allnodenum]["species"] = str(species)
        else:
            spectxt = species.pretty_print(
                show_material=pp_show_material, show_compartment=pp_show_compartments)
            CRNgraph.nodes[allnodenum]["species"] = spectxt
        CRNgraph.nodes[allnodenum]["color"] = species_color
        allnodenum += 1
    # reactions follow, allnodenum is not reset between these two loops
    for rxn in CRN.reactions:
        CRNgraph.add_node(allnodenum)
        CRNgraph.nodes[allnodenum]["type"] = str(rxn.propensity_type)
        if isinstance(rxn.propensity_type, MassAction):
            CRNgraph.nodes[allnodenum]["k"] = str(
                rxn.propensity_type.k_forward)
            CRNgraph.nodes[allnodenum]["k_r"] = str(
                rxn.propensity_type.k_reverse)
        elif hasattr(rxn.propensity_type, "k"):
            CRNgraph.nodes[allnodenum]["k"] = str(rxn.propensity_type.k)
            CRNgraph.nodes[allnodenum]["k_r"] = ''
        else:
            CRNgraph.nodes[allnodenum]["k"] = ''
            CRNgraph.nodes[allnodenum]["k_r"] = ''

        reaction_color = member_dictionary_search(rxn,reactioncolordict)
        if(reaction_color is None):
            reaction_color = default_reaction_color
        # CRNgraph.nodes[allnodenum]
        if isinstance(rxn.propensity_type, MassAction):
            kval = rxn.propensity_type.k_forward
            CRNgraph.nodes[allnodenum]["k"] = str(kval)
        elif hasattr(rxn.propensity_type, "k"):
            kval = rxn.propensity_type.k
            CRNgraph.nodes[allnodenum]["k"] = str(rxn.propensity_type.k)
        else:
            CRNgraph.nodes[allnodenum]["k"] = ''

        if(not useweights):
            kval = 1
        if isinstance(rxn.propensity_type, MassAction):
            krev_val = rxn.propensity_type.k_reverse
        else:
            krev_val = None
        if((krev_val is not None) and (not useweights)):
            krev_val = 1
        for reactant in rxn.inputs:
            #Set edge color
            edge_color = member_dictionary_search(rxn,reactioncolordict)
            if edge_color is None:
                edge_color = default_edge_color
            CRNgraph.add_edge(nodedict[reactant.species],allnodenum,weight=kval)
            CRNgraph.edges[nodedict[reactant.species],allnodenum]['color'] = edge_color
            if(krev_val is not None):
                # if the k is 0 then the node does not exist, right?
                CRNgraph.add_edge(
                    allnodenum, nodedict[reactant.species], weight=krev_val)
                CRNgraph.edges[allnodenum, nodedict[reactant.species]]['color'] = edge_color
        for product in rxn.outputs:
            #TODO species cannot find another species in the nodedict????
            CRNgraph.add_edge(allnodenum,nodedict[product.species],weight=kval)
            CRNgraph.edges[allnodenum,nodedict[product.species]]['color'] = edge_color
            if(krev_val is not None):
                CRNgraph.add_edge(
                    nodedict[product.species], allnodenum, weight=krev_val)
                CRNgraph.edges[nodedict[product.species], allnodenum]['color'] = edge_color
        if(len(rxn.outputs) == 0):
            # this adds an edge to the "nothing" node we made in the beginning
            CRNgraph.add_edge(allnodenum, 0, weight=kval)
            CRNgraph.edges[allnodenum, 0]['color'] = edge_color
            if(krev_val is not None):
                CRNgraph.add_edge(0, allnodenum, weight=krev_val)
                CRNgraph.edges[0, allnodenum]['color'] = edge_color
        elif(len(rxn.inputs) == 0):
            # this adds an edge from the "nothing" node we made in the beginning
            CRNgraph.add_edge(0, allnodenum, weight=kval)
            CRNgraph.edges[0, allnodenum]['color'] = edge_color
            if(krev_val is not None):
                CRNgraph.add_edge(allnodenum, 0, weight=krev_val)
                CRNgraph.edges[allnodenum, 0]['color'] = edge_color
        
        CRNgraph.nodes[allnodenum]["color"] = reaction_color
        if(not use_pretty_print):
            CRNgraph.nodes[allnodenum]["species"] = str(rxn)
        else:
            rxntxt = rxn.pretty_print(
                show_material=pp_show_material, show_rates=pp_show_rates, show_attributes=pp_show_attributes)
            # this will show up as "reaction" in the tool tip
            CRNgraph.nodes[allnodenum]["species"] = rxntxt
        # the name of the reaction is the string representation
        rxnlist += [allnodenum]
        allnodenum += 1
    CRNspeciesonly = CRNgraph.copy()
    CRNspeciesonly.remove_nodes_from(rxnlist)
    CRNreactionsonly = CRNgraph.copy()
    CRNreactionsonly.remove_nodes_from(range(rxnlist[0]))
    return CRNgraph, CRNspeciesonly, CRNreactionsonly


class CRNPlotter:
    class MultiPart:
        def __init__(self,name,parts_list,bound = None):
            """multiple simple parts which are treated as one"""
            self.name = name
            self.parts_list = parts_list
            self.bound = None
        def get_directed(self,direction,bound=None,non_binder=None):
            """returns a copy of itself with the direction changed to the value of 'direction'.
            In the case of MultiPart it also means reversing the order of the subparts.
            A MultiPart binds to things differently from a normal part. the binding is distributed among
            the subparts. "non_binder" indicates a dpl_type which should not be shown binding to things."""
            if(non_binder is None):
                #by default we assume that promoters that are part of MultiParts don't bind to things
                #that's because MultiPart is currently only used for repressible Promoters with Operators
                non_binder = ["Promoter"]
            new_multipart = copy.deepcopy(self)
            bound_for_distribution = None
            if(bound is not None):
                bound_for_distribution = copy.copy(bound)
            elif(self.bound is not None):
                bound_for_distribution = copy.copy(self.bound)
            if(bound_for_distribution is not None):
                #distribute the "bound" items among the parts contained within the MultiPart
                recursion = 10
                while(len(bound_for_distribution)>0 and recursion > 0):
                    for part in new_multipart.parts_list:
                        if(part.dpl_type not in non_binder):
                            if(part.bound is None):
                                part.bound = [bound_for_distribution.pop(0)]
                            else:
                                part.bound += [bound_for_distribution.pop(0)]
                            if(len(bound_for_distribution) == 0):
                                break
                    recursion -= 1
                if(recursion == 0):
                    #this is likely to happen if everything in the multipart is marked as "non binding"
                    raise ValueError(f"reached maximum recursion when trying to populate multipart {self}")
            #actually changing the direction of each part inside the MultiPart
            new_multipart.parts_list = [a.get_directed(direction) for a in new_multipart.parts_list]
            if(direction=="reverse"):
                new_multipart.parts_list = new_multipart.parts_list[::-1]
                return new_multipart
            else:
                return new_multipart
        def get_dpl(self):
            """dnaplotlib takes these dictionaries as input. So this converts the MultiPart object
            into a list of dictionaries that matplotlib can understand"""
            outlist = []
            for part in self.parts_list:
                outlist+= part.get_dpl()
            return outlist
        def __repr__(self):
            return "MultiPart("+",".join([str(a) for a in self.parts_list])+")"
          
    class SimpleConstruct:
        def __init__(self,name,parts_list,circular=False,material_type="dna",label_size=13,added_opts = None):
            """this is a simplified version of a DNAconstruct which mostly only has information relevant to plotting"""
            self.name = name
            self.parts_list = parts_list
            self.circular = circular
            self.material_type = material_type
            self.label_size = label_size
            if(added_opts is None):
                self.added_opts = {}
        def get_dpl(self):
            """output a list of dictionaries suitable for dnaplotlib"""
            outlist = []
            for part in self.parts_list:
                part_dpl = part.get_dpl()
                for subpart_dpl in part_dpl:
                    subpart_dpl["opts"].update(self.added_opts)
                outlist+= part_dpl
            return outlist
        def get_dpl_binders(self):
            """output a dnaplotlib dictionary list to represent the "binders". 
            Binders are "regulation" arcs modified to draw a SBOL glyph instead of a line. """
            my_dpl_output = self.get_dpl()
            out_regs = []
            for design in my_dpl_output:
                if('make_binders' in design):
                    for binder in design['make_binders']:
                        linecolor = 'blue'
                        if(binder.material_type=='dna'):
                            linecolor = 'black'
                        elif(binder.material_type=='rna'):
                            linecolor = 'red'
                        if(hasattr(binder,"color")):
                            bindcolor = binder.color
                        else:
                            bindcolor = (0.5,0.5,0.5)
                        out_reg = {'type':'Binding', 'from_part':design, 'to_part':design,
                                            'opts':{'label':binder.name,'label_size':self.label_size*.7,\
                                                    'color':linecolor, 'label_x_offset':-1,'y_offset':10,\
                                                    'face_color':bindcolor}
                                  }
                        out_reg['opts'].update(binder.added_opts)
                        out_regs+=[out_reg]
            return out_regs,my_dpl_output
        def renderDNA(self,  dna_renderer,ax=None,plot_backbone=True):
            
            part_renderers = dna_renderer.SBOL_part_renderers()
            reg_renderers = dna_renderer.std_reg_renderers()
            if(ax is None):
                figsize = (1,1)
                #fig,ax = plt.subplots(constrained_layout=True)
                fig = plt.figure(figsize=figsize)
                ax = fig.add_axes([0,0,1,1])
                plt.tight_layout(pad=0.0001)
            
            my_regs,my_designs = self.get_dpl_binders()
            for part in my_designs:
                part['opts'].update({'edgecolor':dna_renderer.linecolor})
            start,end = dna_renderer.renderDNA(ax,my_designs,part_renderers,circular=self.circular,\
                                    regs=my_regs, reg_renderers=reg_renderers,plot_backbone=plot_backbone)
            
            fig = ax.get_figure()
            ax.axis('off')
            ''
            ylimits = [None,None]
            xlimits = [None,None]
            relevant_stuff = ax.patches+ax.texts
            for patch in relevant_stuff:
                bbox = patch.get_window_extent(renderer=fig.canvas.get_renderer())
                if(ylimits == [None,None]):
                    ylimits = [bbox.ymin,bbox.ymax]
                if(bbox.ymax > ylimits[1]):
                    ylimits[1] = bbox.ymax
                if(bbox.ymin < ylimits[0]):
                    ylimits[0] = bbox.ymin
                if(xlimits == [None,None]):
                    xlimits = [bbox.xmin,bbox.xmax]
                if(bbox.xmax > xlimits[1]):
                    xlimits[1] = bbox.xmax
                if(bbox.xmin < xlimits[0]):
                    xlimits[0] = bbox.xmin
            xlimits[0],ylimits[0] = fig.transFigure.inverted().transform((xlimits[0], ylimits[0]))

            xlimits[1],ylimits[1] = fig.transFigure.inverted().transform((xlimits[1], ylimits[1]))
            ax.relim()
            yheight = ylimits[1]-ylimits[0]
            xheight = xlimits[1]-xlimits[0]
            fig.set_size_inches(xheight/24,yheight/24)
            ax.set_aspect('equal')
            ax.autoscale_view()
            return ax
    
    class SimplePart:
        def __init__(self,name,dpl_type,direction='forward',bound=None,color=None,\
                        color2=None,show_label=True,label_size = 13,label_y_offset = -8,added_opts=None,material_type=None):
            """a simple 'part' for the sole purpose of rendering using DNAplotlib"""
            self.name = name
            self.color = color
            self.color2 = color2 
            self.dpl_type = dpl_type #this is a string which dnaplotlib knows about
            self.direction=direction #"forward" or "reverse"
            self.bound = bound #this should be a list of SimpleParts
            if(added_opts is None):
                self.added_opts = {}
            else:
                self.added_opts = added_opts #dictionary of keywords for dnaplotlib
            self.show_label =show_label #if the label should be added to the 'opts' upon output
            self.label_size = label_size #font size of the label
            self.label_y_offset = label_y_offset
            self.material_type = material_type
        def get_directed(self,direction,bound=None):
            copied_part = copy.copy(self)
            copied_part.direction = direction
            if(bound is not None):
                copied_part.bound=bound
            elif(self.bound is not None):
                copied_part.bound = self.bound
            return copied_part
        def get_dpl(self,bound=None):
            direction = True
            if(self.direction == "reverse"):
                direction = False
            dpl_out = {'type':self.dpl_type, 'name':self.name, 'fwd':direction}
            opts = {'color':self.color,'color2':self.color2}
            if(self.show_label):
                opts['label']=self.name
                opts['label_size']=self.label_size
                opts['label_y_offset']=self.label_y_offset
            opts.update(self.added_opts)
            dpl_out['opts']=opts
            if(bound is not None):
                dpl_out['make_binders']=bound
            elif(self.bound is not None):
                dpl_out['make_binders']=self.bound

            return [dpl_out]
        def __repr__(self):
            return "SimplePart("+str(self.name)+"-"+str(self.direction)[0]+")"

    def __init__(self,dna_renderer=None,rna_renderer=None,cmap = "Set3",colordict=None):
        if(dna_renderer is None):
            self.dna_renderer=dpl.DNARenderer(scale = 5,linewidth=3)
        else:
            self.dna_renderer = dna_renderer
        if(rna_renderer is None):
            self.rna_renderer=dpl.DNARenderer(scale = 5,linewidth=3,linecolor=(1,0,0))
        else:
            self.rna_renderer = rna_renderer
        if(colordict is None):
            colordict = {}
        self.colordict = colordict
        if(isinstance(cmap,str)):
            self.cmap = plt.get_cmap(cmap).colors
        elif(isinstance(cmap,list)):
            self.cmap = cmap
        else:
            self.cmap = None
        self.color_counter = 0
        self.clear_dicts()
    def renderMixture(self,mixture,crn = None,rna_renderer=None,dna_renderer=None,store=True,output=None,recursion_depth = 4,compiled_components=None):
        """creates dnaplotlib images for all relevant species in a mixture"""
        if(crn is None):
            mycrn,compiled_components = mixture.compile_crn(return_enumerated_components = True,
                                         initial_concentrations_at_end = True,
                                         copy_objects = False,
                                         add_reaction_species = False,recursion_depth=recursion_depth)
        else:
            mycrn = crn
        if(rna_renderer is None and self.rna_renderer is not None):
            rna_renderer = self.rna_renderer
        else:
            raise ValueError("rna_renderer cannot be None")
        if(dna_renderer is None and self.dna_renderer is not None):
            dna_renderer = self.dna_renderer
        else:
            raise ValueError("dna_renderer cannot be None")
        self.clear_dicts()
        if(compiled_components is None):
            #this only happens if CRN and mixture are both given
            _,compiled_components =  mixture.compile_crn(return_enumerated_components = True,
                                            initial_concentrations_at_end = True,
                                            copy_objects = False,
                                            add_reaction_species = False,recursion_depth=recursion_depth)
        for component in compiled_components:
            if(isinstance(component,Construct)):
                a = self.make_dpls_from_construct(component)
        plt.ioff()
        for species in mycrn.species:
            a = self.make_dpl_from_species(species)
            if(a.material_type is not None):
                plot_bb = True
                if(not isinstance(a,self.SimpleConstruct)):
                    plot_bb = False
                    #if we aren't looking at a construct then don't plot the backbone
                    newcon = self.SimpleConstruct(a.name,[a],material_type=a.material_type)
                else:
                    newcon = a
                
                if(newcon.material_type=='rna'):
                    ax = newcon.renderDNA(rna_renderer,plot_backbone=plot_bb)
                else:
                    ax = newcon.renderDNA(dna_renderer,plot_backbone=plot_bb)
                if(store):
                    #this part converts the matplotlibplot into a base64-encoded image
                    imagestream = io.BytesIO()
                    fig = ax.get_figure()
                    if(output is not None):
                        fig.savefig(output+"_"+str(species).replace("_","").replace("forward","f").replace("reverse","r").replace("part","")+".pdf",bbox_inches='tight')
                    fig.savefig(imagestream,bbox_inches='tight')
                    plt.close()
                    png_str = base64.b64encode(imagestream.getvalue())
                    self.species_image_dict[species]= png_str
        return self.species_image_dict
    def renderConstruct(self,construct_obj,rna_renderer=None,dna_renderer=None,showlabels=True, render_rna=False):
        if(rna_renderer is None and self.rna_renderer is not None):
            rna_renderer = self.rna_renderer
        else:
            raise ValueError("rna_renderer cannot be None")
        if(dna_renderer is None and self.dna_renderer is not None):
            dna_renderer = self.dna_renderer
        else:
            raise ValueError("dna_renderer cannot be None")
        a = self.make_dpls_from_construct(construct_obj)
        outaxs = [a.renderDNA(dna_renderer)]
        if(render_rna):
            components = construct_obj.enumerate_constructs()
            for component in components:
                if(isinstance(component,RNA)):
                    a = self.make_dpls_from_construct(component)
                    outaxs += [a.renderDNA(rna_renderer)]
        return outaxs
    def clear_dicts(self):
        self.part_dpl_dict = {}
        self.construct_dpl_dict = {}
        self.species_dpl_dict = {}
        self.species_image_dict ={}
    def get_color(self):
        if(self.cmap is not None):
            out_color = self.cmap[self.color_counter]
            self.color_counter+=1
            if(self.color_counter>=len(self.cmap)):
                self.color_counter = 0
            return out_color
        else:
            raise ValueError("No colormap set")
    def make_dpls_from_construct(self,construct,save_in_dict=True):
        if(construct in self.construct_dpl_dict):
            return self.construct_dpl_dict[construct]
        else:
            new_parts_list = []
            for part in construct:
                new_parts_list += [self.make_dpl_from_part(part,save_in_dict=save_in_dict)]
            
            if(isinstance(construct,DNA)):
                mat_type = "dna"
            elif(isinstance(construct,RNA)):
                mat_type = "rna"
            simple_construct = self.SimpleConstruct(name = construct.name,\
                                                    parts_list=new_parts_list,\
                                                    circular=construct.circular,\
                                                    material_type = mat_type)
            if(save_in_dict):
                self.construct_dpl_dict[construct]=simple_construct
            return simple_construct
    
    def make_dpl_from_species(self,species):
        if(species in self.species_dpl_dict):
            return self.species_dpl_dict[species]
        else:
            if(isinstance(species,OrderedPolymer)):
                #we are dealing with a polymer
                polylist = [] #accumulating the list of SimpleParts
                for monomer in species:
                    #going through the monomers
                    removed_monomer = monomer.get_removed()
                    if(removed_monomer in self.species_dpl_dict):
                        #if we already know about this monomer, just use that
                        polylist += [self.species_dpl_dict[removed_monomer].get_directed(monomer.direction)]
                    elif(isinstance(monomer,ComplexSpecies)):
                        #if the monomer is a complex that means we have to make a bound simplepart
                        binders = []
                        base_simplepart = None #we must figure out who the base is. This is how we do that
                        for specie in monomer.get_species(recursive=True):
                            if(isinstance(specie,ComplexSpecies)):
                                continue
                            if(specie.material_type=='part'):
                                #this material type is only made by dna constructs
                                base_simplepart = copy.copy(self.make_dpl_from_species(specie)) #copy it because now we make it bound to stuff
                                #ideally you already ran make_dpl_from_construct and so this will be populated
                            else:
                                binders += [self.make_dpl_from_species(specie)] #other stuff gets added as a binder
                        base_simplepart.bound = binders
                        self.species_dpl_dict[removed_monomer]=base_simplepart
                        polylist += [base_simplepart.get_directed(monomer.direction)]
                    elif(isinstance(monomer,Species)):
                        #in this case we couldnt find the monomer in our dictionary, but it is
                        #at the base level
                        base_simplepart = self.SimplePart(monomer.name,'UserDefined',color=self.get_color()) #new simplepart
                        self.species_dpl_dict[removed_monomer]=base_simplepart #adding it to the dictionary for the future
                        polylist += [base_simplepart.get_directed(monomer.direction)] #add it to the polymer, with a direction
                out_dpl = self.SimpleConstruct(name = species.name,parts_list = polylist,circular=species.circular,material_type = species.material_type)
            elif(isinstance(species,ComplexSpecies)):
                #if we have a complex but it is not a polymer, we just do the "binding" determination of the polymer part
                base_simplepart = None
                binders = []
                for specie in species.species:
                    #i don't really care who is the base of this thing
                    if(base_simplepart is None):
                        base_simplepart = self.make_dpl_from_species(specie)
                    else:
                        binders += [self.make_dpl_from_species(specie)]
                
                out_dpl = base_simplepart.get_directed(None,binders)
            elif(isinstance(species,Species)):
                #this happens if we have another type of species
                dpl_type = 'Origin' #default is 'origin', which is just a circle
                if(species.material_type=='dna'):
                    dpl_type = 'UserDefined'
                elif(species.material_type=='rna'):
                    dpl_type = 'NCRNA'
                out_dpl = self.SimplePart(name = species.name,dpl_type = dpl_type,color=self.get_color(),material_type=species.material_type)
            
            self.species_dpl_dict[species] = out_dpl
            return out_dpl
    def make_dpl_from_part(self,part,set_color=None,save_in_dict=True,set_color2=None):
        """create a DNAplotlib dictionary from a part"""
        removed_part = part.get_removed()
        if(len(self.colordict)>0 and set_color is None):
            search_color = member_dictionary_search(removed_part,self.colordict)
            if(search_color is not None):
                set_color = search_color
        if(removed_part in self.part_dpl_dict):
            return self.part_dpl_dict[removed_part].get_directed(part.direction)
        else:
            dpl_type= "UserDefined"
            needs_color2 = False
            regs = None
            if(isinstance(part,Promoter)):
                dpl_type = "Promoter"
                if(hasattr(part,"regulators")):
                    regs = part.regulators
            elif(isinstance(part,RBS)):
                dpl_type = "RBS"
            elif(isinstance(part,CDS)):
                dpl_type = "CDS"
            elif(isinstance(part,Protein)):
                dpl_type = "CDS"
            elif(isinstance(part,Operator)):
                dpl_type="Operator"
            elif(isinstance(part,Origin)):
                dpl_type = "Origin"
            elif(isinstance(part,UserDefined)):
                if(part.dpl_type is not None):
                    dpl_type = part.dpl_type
            elif(isinstance(part,Terminator)):
                dpl_type = "Terminator"
            elif(isinstance(part,IntegraseSite)):
                if(part.site_type in ["attP","attB"]):
                    dpl_type = "RecombinaseSite"
                elif(part.site_type in ["attL","attR"]):
                    dpl_type = "RecombinaseSite2"
                    needs_color2 = True
                for key_part in self.part_dpl_dict:
                    stored_part = self.part_dpl_dict[key_part]
                    if(isinstance(key_part,IntegraseSite) and \
                        key_part.integrase==part.integrase and \
                            key_part.dinucleotide==part.dinucleotide):
                        types_list = [key_part.site_type,part.site_type]
                        if(types_list in [["attP","attL"],["attL","attR"],["attR","attL"], ["attB","attR"]]):
                            if(set_color2 is None):
                                set_color2 = stored_part.color
                        if(types_list in [["attB","attL"],["attP","attR"],["attL","attB"],\
                                                                                ["attR","attP"]]):
                            if(set_color is None):
                                set_color = stored_part.color
                        elif(types_list in [["attL","attR"],["attR","attL"],\
                                                                ["attL","attP"],["attR","attB"]]):
                            if(set_color is None):
                                set_color = stored_part.color2
                    if(set_color is not None and set_color2 is not None):
                        #we found all the needed colors so give up
                        break
            if(set_color is None):
                color = self.get_color()
            else:
                color = set_color
            color2 = None
            if(set_color2 is not None):
                color2 = set_color2
            elif(set_color2 is None and needs_color2):
                color2 = self.get_color()
            
            
            outpart = self.SimplePart(name=part.name,\
                                      dpl_type=dpl_type,
                                      color=color,
                                      color2=color2)
            if(regs is not None):
                regparts = []
                for reg in regs:
                    regpart = self.SimplePart(name=reg,dpl_type="Operator",color=color,show_label=False)
                    regparts += [regpart]
                retpart = self.MultiPart(name=part.name,parts_list =[outpart]+regparts)
            else:
                retpart = outpart
            if(save_in_dict):
                self.part_dpl_dict[removed_part] = retpart
                self.species_dpl_dict[part.dna_species] = retpart
            return retpart.get_directed(part.direction)

def render_constructs(constructs,color_dictionary=None):
    """wrapper around CRNPlotter class to make a bunch of constructs which are color coordinated"""
    plotter = CRNPlotter(colordict=color_dictionary)
    axes = []
    for construct in constructs:
        axes += [plotter.renderConstruct(construct)]
    return axes

def render_mixture(mixture,crn,color_dictionary=None,output = None,compiled_components=None):
    plotter = CRNPlotter(colordict=color_dictionary)
    return plotter.renderMixture(mixture,crn,output=output,compiled_components=compiled_components)
def render_network_bokeh(CRN,layout="force", layoutfunc = None, plot_reactions = True, plot_species = True, plot_edges = True, plot_arrows = True,
                        iterations=2000,rseed=30,posscale=1,export=False,species_glyph_size = 12, reaction_glyph_size = 8, export_name = None, **keywords):
    DG, DGspec, DGrxn = generate_networkx_graph(CRN,**keywords) #this creates the networkx objects
    plot = Plot(width=500, height=500, x_range=Range1d(-500, 500), y_range=Range1d(-500, 500)) #this generates a
    show_im = False
    images = None
    if("imagedict" in keywords and keywords["imagedict"] is not None):
        show_im = True
    if(export):
        plot.output_backend = "svg"
    graphPlot(DG,DGspec,DGrxn,plot,layout=layout,posscale=posscale,iterations=iterations,rseed=rseed,show_species_images=show_im,
        layoutfunc=layoutfunc, plot_reactions = plot_reactions, plot_species = plot_species, plot_edges = plot_edges, plot_arrows = plot_arrows,
        species_glyph_size = species_glyph_size, reaction_glyph_size = reaction_glyph_size) 
    if(export):
        if export_name is None:
            raise ValueError("To export you must supply export_name keyword.")
        export_svgs(plot,filename=export_name+".svg")
    return plot
